import numpy as np

from pymoo.core.evaluator import Evaluator
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.core.solution import Solution, SolutionSet
from pymoo.problems.meta import MetaProblem


class NumericalDifferentiation(MetaProblem):

    def __init__(self, problem):
        super().__init__(problem)

    def do(self, X, out, *args, **kwargs):
        self.problem.do(X, out, *args, **kwargs)

        if "dF" in out:
            out["dF"] = NumericalDifferentiationUtil().do(self.problem, X, out["F"], return_values_of=["dF"])


EPS = np.finfo(float).eps


class NumericalDifferentiationUtil:

    def __init__(self, eps=None, jacobian=None, hessian=None):
        """

        eps as in https://stackoverflow.com/questions/31206443/numpy-second-derivative-of-a-ndimensional-array

        Parameters
        ----------
        """
        super().__init__()
        self.eps = eps

        self.jacobian = jacobian
        self.hessian = hessian

        if self.jacobian is None:
            self.jacobian = CentralJacobian()

        if self.hessian is None:
            self.hessian = CentralHessian()

    def do(self, problem, X, F=None, G=None, evaluator=None, hessian=False, return_values_of="auto"):

        # if no evaluator is provided to count function evaluations just use a plain one
        if evaluator is None:
            evaluator = Evaluator()

        # if it is not a population object yet, make one out of it
        pop = to_solution_set(X, F, G)

        # loop over each solution the approximation should be done for
        for solution in pop:
            x = solution.X

            # make sure the solution is evaluated
            if solution.F is None:
                evaluator.eval(problem, solution)

            eps = self.eps
            if eps is None:
                eps = EPS ** (1 / self.jacobian.n_points)
                eps = eps * np.maximum(eps, np.abs(x))

            if not isinstance(eps, np.ndarray):
                eps = np.full(len(x), eps)

            values = ["F"]
            if problem.has_constraints():
                values.append("G")

            self.calc(problem, evaluator, solution, values, eps, hessian)

        if return_values_of == "auto":
            return_values_of = ["dF"]
            if hessian:
                return_values_of.append("ddF")
            if problem.has_constraints():
                return_values_of.append("dG")
                if hessian:
                    return_values_of.append("ddG")

        if isinstance(X, Individual):
            pop = pop[0]

        ret = tuple([pop.get(e) for e in return_values_of])

        if len(ret) == 1:
            return ret[0]
        else:
            return ret

    def calc(self, problem, evaluator, solution, values, eps, hessian):
        jacobian_estm, hessian_estm = self.jacobian, self.hessian

        jac_approx = Population.new(X=jacobian_estm.points(solution.X, eps))
        evaluator.eval(problem, jac_approx)

        for value in values:
            f, F = solution.get(value), jac_approx.get(value)
            dF = np.array([jacobian_estm.calc(f[m], F[:, m], eps) for m in range(F.shape[1])])
            solution.set("d" + value, dF)

        # if the hessian should be calculated as well
        if hessian:
            hess_approx = Population.new(X=self.hessian.points(solution.X, eps))
            evaluator.eval(problem, hess_approx)

            for value in values:
                f, F, FF = solution.get(value), jac_approx.get(value), hess_approx.get(value)
                ddF = np.array([hessian_estm.calc(f[m], F[:, m], FF[:, m], eps) for m in range(FF.shape[1])])
                solution.set("dd" + value, ddF)


# ---------------------------------------------------------------------------------------------------------
# Numerical Differentiation Estimators
# ---------------------------------------------------------------------------------------------------------


class NumericalDifferentiationEstimator:

    def points(self):
        pass

    def calc(self, *args, **kwargs):
        pass


# ---------------------------------------------------------------------------------------------------------
# One Sided Differentiation Estimator
# ---------------------------------------------------------------------------------------------------------


class OneSidedJacobian(NumericalDifferentiationEstimator):
    n_points = 1

    def points(self, x, h):
        return x[None, :].repeat(len(x), axis=0) + np.eye(len(x)) * h

    def calc(self, f, f_jac, h):
        return (f_jac - f) / h


class OneSidedHessian(NumericalDifferentiationEstimator):
    n_points = 1

    def points(self, x, h):
        n = len(x)
        i, j = np.triu_indices(n)
        k = np.arange(len(i))

        eps = np.zeros((len(i), n))
        eps[k, i] += h[i]
        eps[k, j] += h[j]

        return x + eps

    def calc(self, f, f_jac, f_hess, h):
        assert f_jac.ndim == 1 and f_hess.ndim == 1 and len(f_jac) == int(np.sqrt(2 * (len(f_hess))))

        n = len(f_jac)
        i, j = np.triu_indices(n)

        deriv = (f_hess - f_jac[i] - f_jac[j] + f) / (h[i] * h[j])
        hessian = upper_diag_to_sym_full(deriv)

        return hessian


# ---------------------------------------------------------------------------------------------------------
# Central Differentiation Estimator
# ---------------------------------------------------------------------------------------------------------


class CentralJacobian(NumericalDifferentiationEstimator):
    n_points = 2

    def points(self, x, h):
        n = len(x)
        return x + np.row_stack([+1 * np.eye(n) * h, -1 * np.eye(n) * h])

    def calc(self, f, f_jac, h):
        n = int(len(f_jac) / 2)
        return (f_jac[:n] - f_jac[n:]) / (2 * h)


class CentralHessian(NumericalDifferentiationEstimator):
    n_points = 2

    def points(self, x, h):
        n = len(x)

        # [++, --] both having exactly n elements
        eps = [+1 * 2 * np.eye(n) * h, -1 * 2 * np.eye(n) * h]

        i, j = np.triu_indices(n, +1)
        k = np.arange(len(i))

        for s_1 in [+1, -1]:
            for s_2 in [+1, -1]:
                e = np.zeros((len(i), n))
                e[k, i] += s_1 * h[i]
                e[k, j] += s_2 * h[j]
                eps.append(e)

        eps = np.row_stack(eps)
        return x + eps

    def calc(self, f, f_jac, f_hess, h):
        n = int(len(f_jac) / 2)
        i, j = np.triu_indices(n, +1)
        d = np.arange(n)

        pos, neg, mixed = f_hess[:n], f_hess[n:2 * n], f_hess[2 * n:].reshape(4, -1)
        plus_plus, plus_minus, minus_plus, minus_minus = mixed

        hessian = np.zeros((n, n))
        hessian[d, d] = (- pos + 16 * f_jac[:n] - 30 * f + 16 * f_jac[n:] - neg) / (12 * h[d] ** 2)
        hessian[i, j] = (plus_plus - plus_minus - minus_plus + minus_minus) / (4 * h[i] * h[j])
        hessian[np.tril_indices(n, -1)] = hessian[np.triu_indices(n, +1)]

        return hessian


# ---------------------------------------------------------------------------------------------------------
# Complex Step Differentiation Estimator
#
# The Complex-Step Derivative Approximation' by Joaquim R. R. A. MARTINS, Peter STURDZA and Juan J. Alonso published in 2003.
# ---------------------------------------------------------------------------------------------------------


class ComplexStepJacobian(NumericalDifferentiationEstimator):
    n_points = 1

    def points(self, x, h):
        n = len(x)
        return x[None, :].repeat(n, axis=0) + 1j * np.eye(n) * h

    def calc(self, f, f_jac, h):
        return np.imag(f_jac) / h


# ---------------------------------------------------------------------------------------------------------
# Util Methods
# ---------------------------------------------------------------------------------------------------------

def to_solution_set(X, F=None, G=None):
    if isinstance(X, np.ndarray):
        if X.ndim == 1:
            X = Solution(X=X, F=F, G=G)
        else:
            X = Population.new(X=X, F=F, G=G)

    if isinstance(X, Individual):
        X = SolutionSet.create(X)

    return X


def upper_diag_to_sym_full(x):
    n = int(np.sqrt(2 * (len(x))))
    M = np.zeros((n, n))
    M[np.triu_indices(n)] = x
    M[np.tril_indices(n, -1)] = M[np.triu_indices(n, +1)]
    return M



import numpy as np

from pymoo.experimental.benchmarking.analyzer.analyzer import Analyzer
from pymoo.experimental.benchmarking.util import fill_forward_if_nan
from pymoo.indicators.igd import IGD
from pymoo.util.misc import from_dict


class MultiObjectiveAnalyzer(Analyzer):

    def do(self, data, scope=None, benchmark=None, inplace=False, **kwargs):
        assert benchmark is not None, "The benchmark is necessary to retrieve the known optimum of a funtion"

        problem = benchmark.problems[data["problem"]]["obj"]
        CV, F = from_dict(data, "CV", "F")

        igd = np.inf
        pf = problem.pareto_front(**kwargs)
        if pf is not None:
            igd = IGD(pf, zero_to_one=True).do(F)

        ret = {
            "pf": pf,
            "igd": igd,
        }

        if inplace:
            for k, v in ret.items():
                data[k] = v

        return ret


class MultiObjectiveConvergenceAnalyzer(Analyzer):

    def __init__(self, nan_if_not_available=True, **kwargs) -> None:
        super().__init__(**kwargs)
        self.nan_if_not_available = nan_if_not_available

    def do(self, data, **kwargs):
        t = []
        for entry in data:
            t.extend([e['n_evals'] for e in entry["callback"]])
        t = sorted(list(set(t)))

        hash = {}
        for k, v in enumerate(t):
            hash[v] = k

        for i, entry in enumerate(data):

            igd = IGD(entry["pf"], zero_to_one=True)

            vals = np.full(len(t), np.nan)

            for v in entry["callback"]:
                _t, F = from_dict(v, "n_evals", "opt")
                vals[hash[_t]] = igd.do(F)

            fill_forward_if_nan(vals)

            entry["n_evals"] = t
            entry["conv"] = vals



import autograd.numpy as np


from pymoo.core.problem import Problem
from pymoo.problems.autodiff import AutomaticDifferentiation
from pymoo.problems.multi import ZDT, ZDT1


class ElementwiseZDT1(Problem):

    def __init__(self, n_var=30, n_obj=2, n_constr=0, **kwargs):
        super().__init__(n_var, evaluation_of=["F", "dF"], elementwise_evaluation=True, n_obj=n_obj, n_constr=n_constr,
                         **kwargs)

    def _calc_pareto_front(self, n_pareto_points=100):
        x = np.linspace(0, 1, n_pareto_points)
        return np.array([x, 1 - np.sqrt(x)]).T

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = x[0]
        g = 1 + 9.0 / (self.n_var - 1) * x[1:].sum()
        f2 = g * (1 - (f1 / g) ** 0.5)
        out["F"] = ([f1, f2])


class MySphere(Problem):

    def __init__(self, gamma=1):
        super().__init__(n_var=2, n_obj=1, n_constr=1, elementwise_evaluation=True,
                         evaluation_of=["F", "dF", "dG", "ddF", "ddG"])
        self.gamma = gamma

    def _evaluate(self, x, out, *args, **kwargs):
        out["F"] = (x ** 3).sum()
        out["G"] = 1 + (x ** 2).sum()


class ZDT1WithGradient(ZDT):

    def __init__(self, n_var=30, **kwargs):
        super().__init__(n_var, evaluation_of=["F", "dF"], **kwargs)

    def _calc_pareto_front(self, n_pareto_points=100):
        x = np.linspace(0, 1, n_pareto_points)
        return np.array([x, 1 - np.sqrt(x)]).T

    def _evaluate(self, x, out, *args, **kwargs):
        f1 = x[:, 0]
        g = 1 + 9.0 / (self.n_var - 1) * np.sum(x[:, 1:], axis=1)
        f2 = g * (1 - np.power((f1 / g), 0.5))

        out["F"] = np.column_stack([f1, f2])

        if "dF" in out:
            dF = np.zeros([x.shape[0], self.n_obj, self.n_var], dtype=float)
            dF[:, 0, 0], dF[:, 0, 1:] = 1, 0
            dF[:, 1, 0] = -0.5 * np.sqrt(g / x[:, 0])
            dF[:, 1, 1:] = ((9 / (self.n_var - 1)) * (1 - 0.5 * np.sqrt(x[:, 0] / g)))[:, None]
            out["dF"] = dF


if __name__ == '__main__':
    # problem = MySphere()
    # problem = ZDT1WithGradient()

    problem = AutomaticDifferentiation(ZDT1())

    X = np.random.random((10, problem.n_var))

    F, dF = problem.evaluate(X, return_values_of=["F", "dF"])

    print(dF)



from abc import abstractmethod

import autograd.numpy as np

from pymoo.util.cache import Cache
from pymoo.util.misc import at_least_2d_array


# ---------------------------------------------------------------------------------------------------------
# Implementation
# ---------------------------------------------------------------------------------------------------------


class Problem:
    def __init__(self,
                 n_var=-1,
                 n_obj=1,
                 n_constr=0,
                 xl=None,
                 xu=None,
                 check_inconsistencies=True,
                 replace_nan_values_by=np.inf,
                 exclude_from_serialization=None,
                 callback=None,
                 **kwargs):

        """

        Parameters
        ----------
        n_var : int
            Number of Variables

        n_obj : int
            Number of Objectives

        n_constr : int
            Number of Constraints

        xl : np.array, float, int
            Lower bounds for the variables. if integer all lower bounds are equal.

        xu : np.array, float, int
            Upper bounds for the variable. if integer all upper bounds are equal.

        type_var : numpy.dtype


        """

        if "elementwise_evaluation" in kwargs and kwargs.get("elementwise_evaluation"):
            raise Exception("The interface in pymoo 0.5.0 has changed. Please inherit from the ElementwiseProblem "
                            "class AND remove the 'elementwise_evaluation=True' argument to disable this exception.")

        # number of variable
        self.n_var = n_var

        # number of objectives
        self.n_obj = n_obj

        # number of constraints
        self.n_constr = n_constr

        # type of the variable to be evaluated
        self.data = dict(**kwargs)

        # the lower bounds, make sure it is a numpy array with the length of n_var
        self.xl, self.xu = xl, xu

        # a callback function to be called after every evaluation
        self.callback = callback

        # if it is a problem with an actual number of variables - make sure xl and xu are numpy arrays
        if n_var > 0:

            if self.xl is not None:
                if not isinstance(self.xl, np.ndarray):
                    self.xl = np.ones(n_var) * xl
                self.xl = self.xl.astype(float)

            if self.xu is not None:
                if not isinstance(self.xu, np.ndarray):
                    self.xu = np.ones(n_var) * xu
                self.xu = self.xu.astype(float)

        # whether the problem should strictly be checked for inconsistency during evaluation
        self.check_inconsistencies = check_inconsistencies

        # this defines if NaN values should be replaced or not
        self.replace_nan_values_by = replace_nan_values_by

        # attribute which are excluded from being serialized )
        self.exclude_from_serialization = exclude_from_serialization if exclude_from_serialization is not None else []

        # the loader for the pareto set (ps)
        self._pareto_set = Cache(calc_ps, raise_exception=False)

        # the loader for the pareto front (pf)
        self._pareto_front = Cache(calc_pf, raise_exception=False)

        self._ideal_point, self._nadir_point = None, None

    def evaluate(self,
                 X,
                 *args,
                 return_values_of=None,
                 return_as_dictionary=False,
                 **kwargs):

        # make sure the array is at least 2d. store if reshaping was necessary
        X, only_single_value = at_least_2d_array(X, extend_as="row", return_if_reshaped=True)
        assert X.shape[1] == self.n_var, f'Input dimension {X.shape[1]} are not equal to n_var {self.n_var}!'

        # number of function evaluations to be done
        n_evals = X.shape[0]

        # the values to be actually returned by in the end - set bu default if not providded
        ret_vals = default_return_values(self.has_constraints()) if return_values_of is None else return_values_of

        # prepare the dictionary to be filled after the evaluation
        out = dict_with_none(ret_vals)

        # do the actual evaluation for the given problem - calls in _evaluate method internally
        self.do(X, out, *args, **kwargs)

        # make sure the array is 2d before doing the shape check
        out_to_2d_ndarray(out)

        # if enabled (recommended) the output shapes are checked for inconsistencies
        if self.check_inconsistencies:
            check(self, X, out)

        # if the NaN values should be replaced
        if self.replace_nan_values_by is not None:
            replace_nan_values(out, self.replace_nan_values_by)

        # make sure F and G are in fact floats (at least try to do that, no exception will be through if it fails)
        out_to_float(out, ["F", "G"])

        if "CV" in ret_vals or "feasible" in ret_vals:
            CV = calc_constr(out["G"]) if self.has_constraints() else np.zeros([n_evals, 1])
            out["CV"] = CV
            out["feasible"] = CV <= 0

        # in case the input had only one dimension, then remove always the first dimension from each output
        if only_single_value:
            out_to_1d_ndarray(out)

        if self.callback is not None:
            self.callback(X, out)

        # now depending on what should be returned prepare the output
        if return_as_dictionary:
            return out
        else:
            if len(ret_vals) == 1:
                return out[ret_vals[0]]
            else:
                return tuple([out[e] for e in ret_vals])

    def do(self, X, out, *args, **kwargs):
        self._evaluate(X, out, *args, **kwargs)
        out_to_2d_ndarray(out)

    def nadir_point(self):
        """
        Returns
        -------
        nadir_point : np.array
            The nadir point for a multi-objective problem. If single-objective, it returns the best possible solution
            which is equal to the ideal point.

        """
        if self._nadir_point is None:
            pf = self.pareto_front()
            if pf is not None:
                self._nadir_point = np.max(pf, axis=0)
        return self._nadir_point

    def ideal_point(self):
        """
        Returns
        -------
        ideal_point : np.array
            The ideal point for a multi-objective problem. If single-objective it returns the best possible solution.
        """
        if self._ideal_point is None:
            pf = self.pareto_front()
            if pf is not None:
                self._ideal_point = np.min(pf, axis=0)
        return self._ideal_point

    def pareto_front(self, *args, **kwargs):
        return self._pareto_front.exec(self, *args, **kwargs)

    def pareto_set(self, *args, **kwargs):
        return self._pareto_set.exec(self, *args, **kwargs)

    @abstractmethod
    def _evaluate(self, x, out, *args, **kwargs):
        pass

    def has_bounds(self):
        return self.xl is not None and self.xu is not None

    def has_constraints(self):
        return self.n_constr > 0

    def bounds(self):
        return self.xl, self.xu

    def name(self):
        return self.__class__.__name__

    def _calc_pareto_front(self, *args, **kwargs):
        pass

    def _calc_pareto_set(self, *args, **kwargs):
        pass

    @staticmethod
    def calc_constraint_violation(G):
        return calc_constr(G)

    def __str__(self):
        s = "# name: %s\n" % self.name()
        s += "# n_var: %s\n" % self.n_var
        s += "# n_obj: %s\n" % self.n_obj
        s += "# n_constr: %s\n" % self.n_constr
        return s

    def __getstate__(self):
        if self.exclude_from_serialization is not None:
            state = self.__dict__.copy()
            # exclude objects which should not be stored
            for key in self.exclude_from_serialization:
                state[key] = None
            return state
        else:
            return self.__dict__


def calc_ps(problem, *args, **kwargs):
    return at_least_2d_array(problem._calc_pareto_set(*args, **kwargs))


def calc_pf(problem, *args, **kwargs):
    return at_least_2d_array(problem._calc_pareto_front(*args, **kwargs))


# ---------------------------------------------------------------------------------------------------------
# Elementwise Problem
# ---------------------------------------------------------------------------------------------------------


def elementwise_eval(problem, x, out, args, kwargs):
    problem._evaluate(x, out, *args, **kwargs)
    out_to_ndarray(out)
    check(problem, x, out)
    return out


def looped_eval(func_elementwise_eval, problem, X, out, *args, **kwargs):
    return [func_elementwise_eval(problem, x, dict(out), args, kwargs) for x in X]


def starmap_parallelized_eval(func_elementwise_eval, problem, X, out, *args, **kwargs):
    starmap = problem.runner
    params = [(problem, x, dict(out), args, kwargs) for x in X]
    return list(starmap(func_elementwise_eval, params))


def dask_parallelized_eval(func_elementwise_eval, problem, X, out, *args, **kwargs):
    client = problem.runner
    jobs = [client.submit(func_elementwise_eval, problem, x, dict(out), args, kwargs) for x in X]
    return [job.result() for job in jobs]


class ElementwiseProblem(Problem):

    def __init__(self,
                 func_elementwise_eval=elementwise_eval,
                 func_eval=looped_eval,
                 exclude_from_serialization=None,
                 runner=None,
                 **kwargs):

        super().__init__(exclude_from_serialization=exclude_from_serialization, **kwargs)

        # the most granular function which evaluates one single individual - this is the function to parallelize
        self.func_elementwise_eval = func_elementwise_eval

        # the function that calls func_elementwise_eval for ALL solutions to be evaluated
        self.func_eval = func_eval

        # the two ways of parallelization which are supported
        self.runner = runner

        # do not serialize the starmap - this will throw an exception
        self.exclude_from_serialization = self.exclude_from_serialization + ["runner"]

    def do(self, X, out, *args, **kwargs):

        # do an elementwise evaluation and return the results
        ret = self.func_eval(self.func_elementwise_eval, self, X, out, *args, **kwargs)

        # the first element decides what keys will be set
        keys = list(ret[0].keys())

        # now stack all the results for each of them together
        for key in keys:
            assert all([key in _out for _out in ret]), f"For some elements the {key} value has not been set."

            vals = []
            for elem in ret:
                val = elem[key]

                if val is not None:

                    # if it is just a float
                    if isinstance(val, list) or isinstance(val, tuple):
                        val = np.array(val)
                    elif not isinstance(val, np.ndarray):
                        val = np.full(1, val)

                    # otherwise prepare the value to be stacked with each other by extending the dimension
                    val = at_least_2d_array(val, extend_as="row")

                vals.append(val)

            # that means the key has never been set at all
            if all([val is None for val in vals]):
                out[key] = None
            else:
                out[key] = np.row_stack(vals)

        return out

    @abstractmethod
    def _evaluate(self, x, out, *args, **kwargs):
        pass


# ---------------------------------------------------------------------------------------------------------
# Util
# ---------------------------------------------------------------------------------------------------------

def default_return_values(has_constr=False):
    vals = ["F"]
    if has_constr:
        vals.append("CV")
    return vals


def dict_with_none(keys):
    out = {}
    for val in keys:
        out[val] = None
    return out


def out_to_ndarray(out):
    for key, val in out.items():
        if val is not None:
            if not isinstance(val, np.ndarray):
                out[key] = np.array([val])


def out_to_2d_ndarray(out):
    for key, val in out.items():
        if val is not None:
            if isinstance(val, np.ndarray):
                if val.ndim == 1:
                    out[key] = val[:, None]


def out_to_1d_ndarray(out):
    for key in out.keys():
        if out[key] is not None:
            out[key] = out[key][0, :]


def out_to_float(out, keys):
    for key in keys:
        if key in out:
            try:
                out[key] = out[key].astype(float)
            except:
                pass


def calc_constr(G, beta=None, eps=0.0):

    # if G is completely none just return none too - nothing can be inferred
    if G is None:
        return None

    # if the array is empty just return all zeros indicating all are feasible
    elif G.ndim == 1 or G.shape[1] == 0:
        cv = np.zeros((len(G), 1))
        return cv

    else:

        # all values solutions with G less than eps are considered as feasible - 0.0 by default
        G = np.copy(G)
        G[G <= eps] = 0.0

        # sometimes it could be useful to square the constraints for instance and transform the function
        if beta is not None:
            G = (G ** beta)

        # finally sum up the constraints for each solution
        cv = G.sum(axis=1, keepdims=True)

        return cv


def replace_nan_values(out, by=np.inf):
    for key in out:
        try:
            v = out[key]
            v[np.isnan(v)] = by
            out[key] = v
        except:
            pass


def check(problem, X, out):
    elementwise = X.ndim == 1

    # only used if not elementwise
    n_evals = X.shape[0]

    # the values from the output to be checked
    F, dF, G, dG = out.get("F"), out.get("dF"), out.get("G"), out.get("dG")

    # if F is not None:
    #     correct = tuple([problem.n_obj]) if elementwise else (n_evals, problem.n_obj)
    #     assert F.shape == correct, f"Incorrect shape of F: {F.shape} != {correct} (provided != expected)"
    #
    # if dF is not None:
    #     correct = (problem.n_obj, problem.n_var) if elementwise else (n_evals, problem.n_obj, problem.n_var)
    #     assert dF.shape == correct, f"Incorrect shape of dF: {dF.shape} != {correct} (provided != expected)"

    # if G is not None:
    #     if problem.has_constraints():
    #         correct = tuple([problem.n_constr]) if elementwise else (n_evals, problem.n_constr)
    #         assert G.shape == correct, f"Incorrect shape of G: {G.shape} != {correct} (provided != expected)"
    #
    # if dG is not None:
    #     if problem.has_constraints():
    #         correct = (problem.n_constr, problem.n_var) if elementwise else (n_evals, problem.n_constr, problem.n_var)
    #         assert dG.shape == correct, f"Incorrect shape of dG: {dG.shape} != {correct} (provided != expected)"



import random

from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.core.mating import Mating
from pymoo.core.repair import NoRepair
from pymoo.operators.crossover.differental_evolution_crossover import DifferentialEvolutionCrossover
from pymoo.operators.crossover.simulated_binary_crossover import SimulatedBinaryCrossover
from pymoo.operators.mutation.no_mutation import NoMutation
from pymoo.operators.mutation.polynomial_mutation import PolynomialMutation
from pymoo.operators.selection.random_selection import RandomSelection
from pymoo.optimize import minimize
from pymoo.problems.single import Rastrigin


class EnsembleMating(Mating):

    def __init__(self, selections, crossovers, mutations, repairs, **kwargs) -> None:
        super().__init__(None, None, None, **kwargs)
        self.selections = selections
        self.crossovers = crossovers
        self.mutations = mutations
        self.repairs = repairs if len(repairs) > 0 else [NoRepair()]

    def do(self, problem, pop, n_offsprings, **kwargs):
        # randomly choose a combination to be tried
        self.selection = random.choice(self.selections)
        self.crossover = random.choice(self.crossovers)
        self.mutation = random.choice(self.mutations)
        self.repair = random.choice(self.repairs)

        off = super().do(problem, pop, n_offsprings, **kwargs)
        return off


selections = [RandomSelection()]

# define all the crossovers to be tried
crossovers = [SimulatedBinaryCrossover(10.0), SimulatedBinaryCrossover(30.0), DifferentialEvolutionCrossover()]
# COMMENT out this line to only use the SBX crossover with one eta value
# crossovers = [SimulatedBinaryCrossover(30)]

mutations = [NoMutation(), PolynomialMutation(10.0), PolynomialMutation(30.0)]
repairs = []

ensemble = EnsembleMating(selections, crossovers, mutations, repairs)

problem = Rastrigin(n_var=30)

algorithm = GA(
    pop_size=100,
    mating=ensemble,
    eliminate_duplicates=True)

res = minimize(problem,
               algorithm,
               seed=1,
               verbose=True)

print("Best solution found: \nX = %s\nF = %s" % (res.X, res.F))



import numpy as np

from pymoo.experimental.benchmarking.analyzer.analyzer import Analyzer
from pymoo.experimental.benchmarking.util import fill_forward_if_nan


class ConvergenceAnalyzer(Analyzer):

    def __init__(self, nan_if_not_available=True, **kwargs) -> None:
        super().__init__(**kwargs)
        self.nan_if_not_available = nan_if_not_available

    def do(self, data, **kwargs):
        t = []
        for entry in data:
            t.extend([e["n_evals"] for e in entry["callback"]])
        t = sorted(list(set(t)))

        hash = {}
        for k, v in enumerate(t):
            hash[v] = k

        for i, entry in enumerate(data):

            vals = np.full(len(t), np.nan)

            for e in entry["callback"]:
                vals[hash[e["n_evals"]]] = e["f"]

            fill_forward_if_nan(vals)

            entry["n_evals"] = t
            entry["conv"] = vals
            if "fopt" in entry:
                entry["conv_gap"] = vals - entry["fopt"]



import copy


class Individual:

    def __init__(self,
                 X=None, F=None, G=None,
                 dF=None, dG=None,
                 ddF=None, ddG=None,
                 CV=None, feasible=None,
                 **kwargs) -> None:

        # design variables
        self.X = X

        # objectives and constraint values
        self.F = F
        self.G = G

        # first order derivation
        self.dF = dF
        self.dG = dG

        # second order derivation
        self.ddF = ddF
        self.ddG = ddG

        # constraint violation and feasibility flag
        self.CV = CV
        self.feasible = feasible

        # a set storing what has been evaluated
        self.evaluated = set()

        # additional data to be set
        self.data = kwargs
        self.attr = set(self.__dict__.keys())

    def has(self, key):
        return key in self.attr or key in self.data

    def set(self, key, value):
        if key in self.attr:
            self.__dict__[key] = value
        else:
            self.data[key] = value
        return self

    def set_by_dict(self, **kwargs):
        for k, v in kwargs.items():
            self.set(k, v)

    def copy(self, deep=False):
        ind = copy.copy(self)
        ind.data = copy.copy(self.data) if not deep else copy.deepcopy(self.data)
        return ind

    def get(self, *keys):

        def _get(key):
            if key in self.data:
                return self.data[key]
            elif key in self.attr:
                return self.__dict__[key]
            else:
                return None

        ret = []

        for key in keys:
            ret.append(_get(key))

        if len(ret) == 1:
            return ret[0]
        else:
            return tuple(ret)




from pymoo.core.problem import Problem
from pymoo.problems.meta import MetaProblem
from pymoo.util.misc import at_least_2d_array


class ConstraintsAsPenalty(MetaProblem):

    def __init__(self, problem, penalty=1e6):
        super().__init__(problem)
        self.penalty = penalty

        # set the constraints to be zero, because they are now added to the objective
        self.n_constr = 0

    def do(self, x, out, *args, **kwargs):
        self.problem.do(x, out, *args, **kwargs)

        if self.problem.has_constraints():

            F, G = at_least_2d_array(out["F"]), at_least_2d_array(out["G"])
            CV = Problem.calc_constraint_violation(G)

            out["__F__"] = F
            out["__G__"] = G
            out["__CV__"] = CV

            out["F"] = F + self.penalty * CV
            out["G"] = None




import numpy as np

from pymoo.problems.meta import MetaProblem
from pymoo.util.normalization import denormalize, normalize


class ZeroToOne(MetaProblem):

    def __init__(self, problem):
        super().__init__(problem)
        assert self.xl is not None and self.xu is not None, "Both, xl and xu, must be set to redefine the problem!"

        self._xl, self._xu = problem.xl, problem.xu
        self.xl, self.xu = np.zeros(self.n_var), np.ones(self.n_var)

    def do(self, x, out, *args, **kwargs):
        out["__X__"] = x
        _x = (x - self.xl) / (self.xu - self.xl)
        super().do(_x, out, *args, **kwargs)

    def denormalize(self, x):
        return denormalize(x, self._xl, self._xu)

    def normalize(self, x):
        return normalize(x, self._xl, self._xu)





import abc

from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.core.algorithm import Algorithm
from pymoo.core.population import pop_from_array_or_individual
from pymoo.operators.sampling.lhs import LatinHypercubeSampling
from pymoo.util.display import SingleObjectiveDisplay
from pymoo.util.termination.f_tol_single import SingleObjectiveSpaceToleranceTermination


class LocalSearch(Algorithm):

    def __init__(self,
                 x0=None,
                 sampling=LatinHypercubeSampling(),
                 display=SingleObjectiveDisplay(),
                 n_sample_points="auto",
                 n_max_sample_points=20,
                 **kwargs):

        super().__init__(display=display, **kwargs)
        self.sampling = sampling
        self.n_sample_points = n_sample_points
        self.n_max_sample_points = n_max_sample_points
        self.x0 = x0
        self.default_termination = SingleObjectiveSpaceToleranceTermination(n_last=5, tol=1e-8)

        self.is_local_initialized = False

    def _setup(self, problem, x0=None, **kwargs):
        if self.x0 is None:
            self.x0 = x0

    def _initialize_infill(self):
        # calculate the default number of sample points
        if self.n_sample_points == "auto":
            self.n_sample_points = min(self.problem.n_var * 5, self.n_max_sample_points)

        # no initial point is provided - sample in bounds and take the best
        if self.x0 is None:
            if not self.problem.has_bounds():
                raise Exception("Either provide an x0 or a problem with variable bounds!")
            pop = self.sampling.do(self.problem, self.n_sample_points)
        else:
            pop = pop_from_array_or_individual(self.x0)

        return pop

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills=infills, **kwargs)

        self.evaluator.eval(self.problem, infills, algorithm=self)
        self.x0 = FitnessSurvival().do(self.problem, infills, n_survive=1)[0]

    def _infill(self):
        if not self.is_local_initialized:
            return self._local_initialize_infill()
        else:
            return self._local_infill()

    def _advance(self, **kwargs):
        if not self.is_local_initialized:
            self.is_local_initialized = True
            return self._local_initialize_advance(**kwargs)
        else:
            return self._local_advance(**kwargs)

    def _local_initialize_infill(self, *args, **kwargs):
        return self._local_infill(*args, **kwargs)

    def _local_initialize_advance(self, *args, **kwargs):
        return self._local_advance(*args, **kwargs)

    @abc.abstractmethod
    def _local_infill(self):
        pass

    @abc.abstractmethod
    def _local_advance(self, **kwargs):
        pass



import numpy as np
from scipy.spatial.distance import cdist

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.core.population import Population
from pymoo.decomposition.tchebicheff import Tchebicheff
from pymoo.factory import get_decomposition
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.util.display import MultiObjectiveDisplay
from pymoo.util.misc import set_if_none
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting


# =========================================================================================================
# Implementation
# =========================================================================================================


class MOEADAWA(GeneticAlgorithm):
    """

    Adapted from
    Qi, Y., Ma, X., Liu, F., Jiao, L., Sun, J., & Wu, J. (2014). MOEA/D with Adaptive Weight Adjustment.
    Evolutionary Computation, 22(2), 231–264. https://doi.org/10.1162/EVCO_a_00109

    Implementation inspired from the MOEAD-AWA implementation in Matlab of the PlatEMO library.

    """

    def __init__(self,
                 ref_dirs,
                 n_neighbors=20,
                 decomposition=Tchebicheff(),
                 prob_neighbor_mating=0.9,
                 rate_update_weight=0.05,
                 rate_evol=0.8,
                 wag=20,
                 archive_size_multiplier=1.5,
                 use_new_ref_dirs_initialization=True,
                 display=MultiObjectiveDisplay(),
                 **kwargs):
        """

        MOEAD-AWA Algorithm.

        Parameters
        ----------
        ref_dirs
        n_neighbors
        decomposition
        prob_neighbor_mating
        rate_update_weight
        rate_evol
        wag
        archive_size_multiplier
        use_new_ref_dirs_initialization
        display
        kwargs
        """

        self.n_neighbors = n_neighbors
        self.prob_neighbor_mating = prob_neighbor_mating
        self.decomp = decomposition

        self.rate_update_weight = rate_update_weight
        self.rate_evol = rate_evol
        self.wag = wag

        self.EP = None
        self.nEP = np.ceil(len(ref_dirs) * archive_size_multiplier)

        set_if_none(kwargs, 'pop_size', len(ref_dirs))
        set_if_none(kwargs, 'sampling', FloatRandomSampling())
        set_if_none(kwargs, 'crossover', SBX(prob=1.0, eta=20))
        set_if_none(kwargs, 'mutation', PM(prob=None, eta=20))
        set_if_none(kwargs, 'survival', None)
        set_if_none(kwargs, 'selection', None)

        super().__init__(display=display, **kwargs)

        # initialized when problem is known
        self.ref_dirs = ref_dirs
        if use_new_ref_dirs_initialization:
            self.ref_dirs = 1.0 / (self.ref_dirs + 1e-6)
            self.ref_dirs = self.ref_dirs / np.sum(self.ref_dirs, axis=1)[:, None]

        if self.ref_dirs.shape[0] < self.n_neighbors:
            print("Setting number of neighbours to population size: %s" % self.ref_dirs.shape[0])
            self.n_neighbors = self.ref_dirs.shape[0]

        self.nds = NonDominatedSorting()

        # compute neighbors of reference directions using euclidean distance
        self._update_neighbors()

    def _update_neighbors(self):
        self.neighbors = np.argsort(cdist(self.ref_dirs, self.ref_dirs), axis=1, kind='quicksort')[:, :self.n_neighbors]

    def _setup(self, problem, **kwargs):
        assert not problem.has_constraints(), "This implementation of MOEAD-AWA does not support any constraints."

        if isinstance(self.decomp, str):
            # for one or two objectives use tchebi otherwise pbi
            if self.decomp == 'auto':
                if self.problem.n_obj <= 2:
                    from pymoo.decomposition.tchebicheff import Tchebicheff
                    self.decomp = Tchebicheff()
                else:
                    from pymoo.decomposition.pbi import PBI
                    self.decomp = PBI()

    def _initialize_advance(self, infills=None, **kwargs):
        self.ideal_point = np.min(infills.get("F"), axis=0)
        self.EP = Population()

    def _infill(self):
        pass

    def _advance(self, **kwargs):
        repair, crossover, mutation = self.repair, self.mating.crossover, self.mating.mutation

        # retrieve the current population
        pop = self.pop

        offsprings = Population(len(pop))

        # iterate for each member of the population in random order
        for i in np.random.permutation(len(pop)):

            # all neighbors of this individual and corresponding weights
            N = self.neighbors[i, :]

            if np.random.random() < self.prob_neighbor_mating:
                parents = N[np.random.permutation(self.n_neighbors)]
            else:
                parents = np.random.permutation(self.pop_size)

            # do recombination and create an offspring
            off = crossover.do(self.problem, pop, parents[None, :crossover.n_parents], algorithm=self)
            off = mutation.do(self.problem, off, algorithm=self)
            off = off[np.random.randint(0, len(off))]

            # repair first in case it is necessary - disabled if instance of NoRepair
            off = repair.do(self.problem, off, algorithm=self)

            # evaluate the offspring
            self.evaluator.eval(self.problem, off, algorithm=self)

            # update the ideal point
            self.ideal_point = np.min(np.vstack([self.ideal_point, off.F]), axis=0)

            FV = self.decomp.do(pop[parents].get("F"),weights=self.ref_dirs[parents], ideal_point=self.ideal_point)
            off_FV = self.decomp.do(off.F[None, :], weights=self.ref_dirs[parents], ideal_point=self.ideal_point)
            I = np.where(off_FV < FV)[0]  # find locations where the offspring is strictly better than a parent
            pop[parents[I]] = off  # replace these parents by the offspring
            offsprings[i] = off  # save the offspring for the eventual external archive update

        # AWA specific part - update elite archive and reference directions
        # if no max generation is set then the behavior is always AWA
        max_gen = self.termination.__dict__.get("n_max_gen", 0)
        if self.n_gen >= self.rate_evol * max_gen and self.n_gen % self.wag == 0:
            # update external archive that should be bigger than population size
            self.EP = update_ep(self.EP, offsprings, self.nEP, self.nds)

            # remove a fraction of the crowded weights vectors and add new ones using the elite archive EP
            self.pop, self.ref_dirs = update_weight(self.pop,
                                                    self.ref_dirs,
                                                    # the small epsilon prevent NaNs although they won't stay anyway
                                                    self.ideal_point - 10e-7,
                                                    self.EP,
                                                    self.rate_update_weight * len(self.pop))

            # update the neighbor locations, it was not done in the PlatEMO implementation but it is in the paper
            self._update_neighbors()


def update_ep(EP, Offsprings, nEP, nds):
    """
    Update the external population archive
    """

    # merge population and keep only the first non-dominated front
    EP = Population.merge(EP, Offsprings)
    EP = EP[nds.do(EP.get("F"), only_non_dominated_front=True)]
    N, M = EP.get("F").shape

    # Delete the overcrowded solutions
    D = cdist(EP.get("F"), EP.get("F"))

    # prevent selection of a point with itself
    D[np.eye(len(D), dtype=np.bool)] = np.inf

    removed = np.zeros(N, dtype=np.bool)
    while sum(removed) < N - nEP:
        remain = np.flatnonzero(~removed)
        subDis = np.sort(D[np.ix_(remain, remain)], axis=1)

        # compute viscinity distance among the closest neighbors
        prodDist = np.prod(subDis[:, 0: min(M, len(remain))], axis=1)

        # select the point with the smallest viscinity distance to its neighbors and set it as removed
        worst = np.argmin(prodDist)
        removed[remain[worst]] = True

    return EP[~removed]


def update_weight(pop, W, Z, EP, nus):
    """
    Delete crowded weight vectors and add new ones
    """

    N, M = pop.get("F").shape

    # Update the current population by EP
    # Calculate the function value of each solution in Population or EP on each subproblem in W
    combined = Population.merge(pop, EP)
    combinedF = np.abs(combined.get("F") - Z)
    g = np.zeros((len(combined), W.shape[0]))
    for i in range(W.shape[0]):
        g[:, i] = np.max(combinedF * W[i, :], axis=1)

    # Choose the best solution for each subproblem
    best = np.argmin(g, axis=0)
    pop = combined[best]

    ######################################
    # Delete the overcrowded subproblems #
    ######################################

    D = cdist(pop.get("F"), pop.get("F"))

    # avoid selection of solutions with themselves
    D[np.eye(len(D), dtype=np.bool)] = np.inf

    deleted = np.zeros(len(pop), dtype=np.bool)
    while np.sum(deleted) < min(nus, len(EP)):
        remain = np.flatnonzero(~deleted)
        subD = D[np.ix_(remain, remain)]
        subDis = np.sort(subD, axis=1)

        # use viscinity distance to find the most crowded vector among the remaining ones and set it to be deleted
        worst = np.argmin(np.prod(subDis[:, 0:min(M, len(remain))], axis=1))
        deleted[remain[worst]] = True

    pop = pop[~deleted]
    W = W[~deleted, :]

    ######################################
    #       Add new subproblems          #
    ######################################

    # Determine the new solutions be added
    combined = Population.merge(pop, EP)
    selected = np.zeros(len(combined), dtype=np.bool)
    selected[:len(pop)] = True  # keep all solutions from pop and add solutions from EP
    D = cdist(combined.get("F"), combined.get("F"))
    D[np.eye(len(D), dtype=np.bool)] = np.inf
    while np.sum(selected) < min(N, len(selected)):
        # get the farthest solutions from already selected solutions using viscinity distance and select it
        subDis = np.sort(D[np.ix_(~selected, selected)], axis=1)
        best = np.argmax(np.prod(subDis[:, 0:min(M, subDis.shape[1])], axis=1))
        remain = np.flatnonzero(~selected)
        selected[remain[best]] = True

    # Add new subproblems to W
    newF = EP[selected[len(pop):]].get("F")

    # transform the weights using the WS transformation described in the paper
    # we don't care about NaNs as they will be eliminated later anyway
    with np.errstate(divide="ignore", invalid='ignore'):
        temp = 1. / (newF - Z)
        W = np.vstack([W, temp / np.sum(temp, axis=1)[:, None]])

    # Add new solutions
    pop = combined[selected]
    return pop, W



from pymoo.experimental.benchmarking.util import filter_by


class Analyzer:

    def __init__(self, default_group_by=None) -> None:
        super().__init__()
        self.default_group_by = default_group_by

    def run(self, result, group_by=None, **kwargs):

        if group_by is not None:
            groups = list(filter_by(result, group_by, return_group=True))
        else:
            groups = [(e, e) for e in result]

        ret = []
        for scope, data in groups:
            vals = self.do(data, scope=scope, **kwargs)
            if vals is not None:
                ret.append({**scope, **vals})
            else:
                ret.extend(data)

        return ret

    def do(self, scope, data, **kwargs):
        pass


import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.algorithms.soo.nonconvex.de import DES
from pymoo.docs import parse_doc_string
from pymoo.core.survival import Survival
from pymoo.operators.crossover.dex import DEX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.util.display import SingleObjectiveDisplay
from pymoo.util.misc import cdist
from pymoo.util.termination.default import SingleObjectiveDefaultTermination
from pymoo.visualization.video.two_var_one_obj import TwoVariablesOneObjectiveVisualization


# =========================================================================================================
# Survival
# =========================================================================================================


class NichingSurvival(Survival):

    def __init__(self, d1, d2, max_niches, max_crowding) -> None:
        super().__init__(True)
        self.d1 = d1
        self.d2 = d2
        self.max_niches = max_niches
        self.max_crowding = max_crowding

    def _do(self, problem, pop, n_survive=None, out=None, **kwargs):
        F = pop.get("F")
        assert F.shape[1] == 1, "This survival only works for single-objective problems"

        I = np.argsort(F[:, 0])
        pop = pop[I]
        X, F = pop.get("X", "F")

        xl, xu = problem.bounds()
        X = (X - xl) / (xu - xl)

        func_dist = lambda _x, _X: cdist(_x[None, :], _X)[0] / (problem.n_var ** 0.5)

        niches = np.full(len(pop), 0)
        crowding = np.full(len(pop), 0)
        penalty = np.full(len(pop), 0)

        leaders = [0]

        for i in range(1, len(pop)):

            # calculate the distance to all solutions with a better function value than i
            d = func_dist(X[i], X[:i])

            closer_than_d1 = d <= self.d1

            # if the distance is less than d1, then we inherit the niche
            if np.any(closer_than_d1):

                # find the first solution with distance less than d1 - lowest objective value
                J = np.argmax(closer_than_d1)
                niche = niches[J]

            else:

                # create a new niche if maximum has not reached yet
                if len(leaders) < self.max_niches:
                    niche = len(leaders)
                    leaders.append(i)

                # else assign the closest niche
                else:
                    niche = np.argmin(func_dist(X[i], X[leaders]))

            niches[i] = niche

            closer_than_d2 = d <= self.d2

            # if any other solution is in the crowding radius
            if np.any(closer_than_d2):

                # assign the crowding counter for all of them
                J, = np.where(closer_than_d2)
                crowding[J] += 1

                # the penalty is defined by the crowding counter minus the allowed max. allowed crowding
                penalty[i] = max(0, (crowding[J]).max() - self.max_crowding)

        # create the niche assignment matrix A
        A = [[] for _ in range(niches.max() + 1)]
        for i, niche in enumerate(niches):
            A[niche].append(i)

        # set the local optimum rank for each niche
        rank = np.full(len(pop), -1)
        for I in A:
            rank[I] = np.arange(len(I))

        J = np.lexsort((rank, niches, penalty))

        return pop[J[:n_survive]]


# ====================

class LOR2(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 selection=DES("rand"),
                 crossover=DEX(CR=0.7),
                 mutation=PM(eta=20),
                 survival=NichingSurvival(0.1, 0.01, 5, 4),
                 eliminate_duplicates=True,
                 n_offsprings=None,
                 display=SingleObjectiveDisplay(),
                 **kwargs):

        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=survival,
                         eliminate_duplicates=eliminate_duplicates,
                         n_offsprings=n_offsprings,
                         display=display,
                         **kwargs)

        self.default_termination = SingleObjectiveDefaultTermination()

    def _set_optimum(self):
        super()._set_optimum()
        # print(self.opt.get("X"))


parse_doc_string(LOR2.__init__)

from pymoo.factory import get_problem
from pymoo.optimize import minimize

problem = get_problem("himmelblau")

algorithm = LOR2(
    pop_size=100,
    eliminate_duplicates=True,
    callback=TwoVariablesOneObjectiveVisualization()
)

res = minimize(problem,
               algorithm,
               ("n_gen", 10000),
               seed=1,
               verbose=True)

print("Best solution found: \nX = %s\nF = %s" % (res.X, res.F))



import numpy as np
from scipy.spatial.distance import cdist

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.core.duplicate import NoDuplicateElimination
from pymoo.core.population import Population
from pymoo.core.selection import Selection
from pymoo.docs import parse_doc_string
from pymoo.operators.crossover.sbx import SimulatedBinaryCrossover
from pymoo.operators.mutation.pm import PolynomialMutation
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.util.display import MultiObjectiveDisplay


# =========================================================================================================
# Neighborhood Selection
# =========================================================================================================


class NeighborhoodSelection(Selection):

    def __init__(self, neighbors, prob=1.0) -> None:
        super().__init__()
        self.neighbors = neighbors
        self.prob = prob

    def _do(self, pop, n_select, n_parents, k=None, **kwargs):
        if k is None:
            k = np.random.permutation(len(pop))[:n_select]
        assert len(k) == n_select

        N = self.neighbors
        P = np.full((n_select, n_parents), -1)

        for i, j in enumerate(k):
            if np.random.random() < self.prob:
                P[i] = np.random.choice(N[j], n_parents, replace=False)
            else:
                P[i] = np.random.permutation(len(pop))[:n_parents]

        return P


# =========================================================================================================
# Implementation
# =========================================================================================================

class MOEAD(GeneticAlgorithm):

    def __init__(self,
                 ref_dirs,
                 n_neighbors=20,
                 decomposition='auto',
                 prob_neighbor_mating=0.9,
                 sampling=FloatRandomSampling(),
                 crossover=SimulatedBinaryCrossover(prob=1.0, eta=20),
                 mutation=PolynomialMutation(prob=None, eta=20),
                 display=MultiObjectiveDisplay(),
                 **kwargs):
        """
        Parameters
        ----------
        ref_dirs
        n_neighbors
        decomposition
        prob_neighbor_mating
        display
        kwargs
        """

        self.ref_dirs = ref_dirs
        self.n_neighbors = min(len(ref_dirs), n_neighbors)
        self.prob_neighbor_mating = prob_neighbor_mating
        self.decomp = decomposition

        # neighbours includes the entry by itself intentionally for the survival method
        self.neighbors = np.argsort(cdist(self.ref_dirs, self.ref_dirs), axis=1, kind='quicksort')[:, :self.n_neighbors]

        self.selection = NeighborhoodSelection(self.neighbors, prob=prob_neighbor_mating)

        super().__init__(pop_size=len(ref_dirs), sampling=sampling, crossover=crossover, mutation=mutation,
                         eliminate_duplicates=NoDuplicateElimination(), display=display,
                         advance_after_initialization=False, **kwargs)

        # the mating is just performed once here - population does not need to be filled up
        self.mating.n_max_iterations = 1

    def _setup(self, problem, **kwargs):
        assert not problem.has_constraints(), "This implementation of MOEAD does not support any constraints."

        if isinstance(self.decomp, str):
            # for one or two objectives use tchebi otherwise pbi
            if self.decomp == 'auto':
                if self.problem.n_obj <= 2:
                    from pymoo.decomposition.tchebicheff import Tchebicheff
                    self.decomp = Tchebicheff()
                else:
                    from pymoo.decomposition.pbi import PBI
                    self.decomp = PBI()

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills, **kwargs)
        self.ideal = np.min(self.pop.get("F"), axis=0)

    def _infill(self):
        # MOEA\D inherits from genetic algorithm but does not implement the infill/advance interface
        pass

    def _advance(self, **kwargs):
        pop = self.pop

        # iterate for each member of the population in random order
        for i in np.random.permutation(len(pop)):
            # get the parents using the neighborhood selection
            P = self.selection.do(pop, 1, self.mating.crossover.n_parents, k=[i])

            # perform a mating using the default operators - if more than one offspring just pick the first
            off = self.mating.do(self.problem, pop, 1, parents=P)[0]

            # evaluate the offspring
            self.evaluator.eval(self.problem, off, algorithm=self)

            # update the ideal point
            self.ideal = np.min(np.vstack([self.ideal, off.F]), axis=0)

            # now actually do the replacement of the individual is better
            self._replace(i, off)

    def _replace(self, i, off):
        pop = self.pop

        # calculate the decomposed values for each neighbor
        N = self.neighbors[i]
        FV = self.decomp.do(pop[N].get("F"), weights=self.ref_dirs[N, :], ideal_point=self.ideal)
        off_FV = self.decomp.do(off.F[None, :], weights=self.ref_dirs[N, :], ideal_point=self.ideal)

        # this makes the algorithm to support constraints - not originally proposed though and not tested enough
        # if self.problem.has_constraints():
        #     CV, off_CV = pop[N].get("CV")[:, 0], np.full(len(off_FV), off.CV)
        #     fmax = max(FV.max(), off_FV.max())
        #     FV, off_FV = parameter_less(FV, CV, fmax=fmax), parameter_less(off_FV, off_CV, fmax=fmax)

        # get the absolute index in F where offspring is better than the current F (decomposed space)
        I = np.where(off_FV < FV)[0]
        pop[N[I]] = off


class ParallelMOEAD(MOEAD):

    def __init__(self, ref_dirs, **kwargs):
        super().__init__(ref_dirs, **kwargs)
        self.indices = None

    def _infill(self):
        pop_size, cross_parents, cross_off = self.pop_size, self.mating.crossover.n_parents, self.mating.crossover.n_offsprings

        # do the mating in a random order
        I = np.random.permutation(len(self.pop))[:self.n_offsprings]

        # get the parents using the neighborhood selection
        P = self.selection.do(self.pop, self.n_offsprings, cross_parents, k=I)

        # do not any duplicates elimination - thus this results in exactly pop_size * n_offsprings offsprings
        off = self.mating.do(self.problem, self.pop, np.inf, parents=P)

        # select a random offspring from each mating
        off = Population.create(*[np.random.choice(pool) for pool in np.reshape(off, (self.n_offsprings, -1))])

        # store the indices because of the neighborhood matching in advance
        self.indices = I

        return off

    def _advance(self, infills=None, **kwargs):
        assert len(self.indices) == len(infills), "Number of infills must be equal to the one created beforehand."

        # update the ideal point before starting to replace
        self.ideal = np.min(np.vstack([self.ideal, infills.get("F")]), axis=0)

        # now do the replacements as in the loop-wise version
        for i, off in enumerate(infills):
            self._replace(self.indices[i], off)


parse_doc_string(MOEAD.__init__)



from pymoo.experimental.benchmarking.analyzer.analyzer import Analyzer

DEFAULT_FUNCTIONS = {
    "first": lambda x: x[0],
    "collect": lambda x: x,
}


class GroupBy(Analyzer):

    def __init__(self,
                 attrs,
                 funcs={},
                 **kwargs) -> None:
        """
        This is a result analyzer which groups elements and aggregates attributes.
        Similar to database queries but on a Python dictionary level.

        Parameters
        ----------
        attrs : tuple
            A tuple defining (in,func,out) which is an aggregation of one attribute.

        funcs : dict
            If some other functions should be injected to be known for the analysis.

        """
        super().__init__(**kwargs)
        self.attrs = attrs
        self.funcs = {**DEFAULT_FUNCTIONS, **funcs}

    def do(self, data, **kwargs):
        ret = {}

        for elems in self.attrs:

            # parse the attribute inputs
            if len(elems) == 2:
                _in, _func = elems
                _out = f"{str(_func)}({_in})"
            elif len(elems) == 3:
                _in, _func, _out = elems
            else:
                raise Exception("The group by attributes are either (in, func) or (in,func,out).")

            if isinstance(_func, str):
                if _func not in self.funcs:
                    raise Exception(f"Unknown Function: {_func}")
                _func = self.funcs[_func]

            # get the data from the current data set
            vals = [entry[_in] for entry in data]

            # apply the function to the input
            aggr = _func(vals)

            # finally store the output dta
            ret[f"{_out}"] = aggr

        return ret



"""
This class provide an interface for other libraries to specific modules. For example, the evolutionary operations
can be used easily just by calling a function and providing the lower and upper bounds of the problem.

"""

import numpy as np

from pymoo.core.population import Population
from pymoo.core.problem import Problem


# =========================================================================================================
# A global interface for some features
# =========================================================================================================


def get_problem_func(n_var, xl, xu, type_var):
    class P(Problem):
        def __init__(self) -> None:
            super().__init__(n_var=n_var, n_obj=1, n_constr=0, xl=xl, xu=xu, type_var=type_var)

    return P


def sample(sampling, n_samples, n_var, xl=0, xu=1, **kwargs):
    problem = get_problem_func(n_var, xl, xu, None)(**kwargs)
    return sampling.do(problem, n_samples, pop=None, **kwargs)


def crossover(crossover, a, b, c=None, xl=0, xu=1, type_var=np.double, **kwargs):
    n = a.shape[0]
    _pop = Population.merge(Population.new("X", a), Population.new("X", b))
    _P = np.column_stack([np.arange(n), np.arange(n) + n])

    if c is not None:
        _pop = Population.merge(_pop, Population.new("X", c))
        _P = np.column_stack([_P, np.arange(n) + 2 * n])

    problem = get_problem_func(a.shape[1], xl, xu, type_var)(**kwargs)
    return crossover.do(problem, _pop, _P, **kwargs).get("X")


def mutation(mutation, X, xl=0, xu=1, type_var=np.double, **kwargs):
    problem = get_problem_func(X.shape[1], xl, xu, type_var)(**kwargs)
    return mutation.do(problem, Population.new("X", X), **kwargs).get("X")





from pymoo.core.problem import Problem


class MetaProblem(Problem):

    def __init__(self, problem, override_evaluate=False):
        self.__dict__.update(**problem.__dict__)
        self.problem = problem
        self.override_evaluate = override_evaluate

    def do(self, X, out, *args, **kwargs):
        obj = super() if not self.override_evaluate else self.problem
        obj.do(X, out, *args, **kwargs)

    def _evaluate(self, x, out, *args, **kwargs):
        self.problem._evaluate(x, out, *args, **kwargs)

    def _calc_pareto_front(self, *args, **kwargs):
        return self.problem._calc_pareto_front(*args, **kwargs)

    def _calc_pareto_set(self, *args, **kwargs):
        return self.problem._calc_pareto_set(*args, **kwargs)




import numpy as np

from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.util.misc import at_least_2d_array


class Converter:

    def __init__(self, attr="X") -> None:
        super().__init__()
        self.attr = attr
        self.obj = None

    def forward(self, obj, **kwargs):
        self.obj = obj
        return self._forward(obj, **kwargs)

    def backward(self, obj, **kwargs):
        return self._backward(obj, **kwargs)

    def _forward(self, obj, **kwargs):
        return obj

    def _backward(self, obj, **kwargs):
        return obj


class PopulationToNumpy(Converter):

    def _forward(self, pop, **kwargs):
        assert self.attr is not None and isinstance(pop, Population)
        return pop.get(self.attr)

    def _backward(self, X, inplace=False, **kwargs):
        assert isinstance(X, np.ndarray) and X.ndim == 2
        if inplace:
            self.obj.set(self.attr, X)
            return self.obj
        else:
            return Population.new(**{self.attr: X})


class IndividualToNumpy(Converter):

    def _forward(self, ind, **kwargs):
        assert self.attr is not None and isinstance(ind, Individual)
        return at_least_2d_array(ind.get(self.attr), extend_as="row")

    def _backward(self, X, inplace=False, **kwargs):
        assert isinstance(X, np.ndarray)

        # if it should be 2d we convert it to be a 1-d array
        if X.ndim == 2 and len(X) == 1:
            vals = X[0]

        assert vals.ndim == 1

        if inplace:
            self.obj.set(self.attr, vals)
            return self.obj
        else:
            return Individual(**{self.attr: vals})


class OneToTwoDimensionalNumpy(Converter):

    def _forward(self, x, **kwargs):
        assert isinstance(x, np.ndarray)
        if x.ndim == 1:
            return x[None, :]
        elif x.ndim == 2:
            return x
        else:
            raise Exception("Error while converting numpy array to 2d.")

    def _backward(self, x, inplace=False):
        assert isinstance(x, np.ndarray)

        # make sure it is actually 2d
        if x.ndim == 1:
            return x[None, :]
        elif x.ndim == 2:
            pass
        else:
            raise Exception("Error while converting numpy array to 2d.")

        if inplace:
            self.obj[:] = x
            return self.obj
        else:
            return x[0]


def get_converter(val):
    if isinstance(val, Population):
        return PopulationToNumpy()
    elif isinstance(val, Individual):
        return IndividualToNumpy()
    elif isinstance(val, np.ndarray):
        if val.ndim == 1:
            return OneToTwoDimensionalNumpy()
        elif val.ndim == 2:
            return Converter()
        else:
            raise Exception("No converter for numpy arrays with dimensions more than 2 exists.")


class DefaultConverter(Converter):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.converter = None

    def forward(self, obj, **kwargs):
        self.converter = get_converter(obj)
        return self.converter.forward(obj, **kwargs)

    def backward(self, obj, **kwargs):
        return self.converter.backward(obj, **kwargs)



from os.path import dirname, realpath

from pymoo.version import __version__


class Config:
    """
    The configuration of this package in general providing the place
    for declaring global variables.
    """

    # the root directory where the package is located at
    root = dirname(realpath(__file__))

    # whether a warning should be printed if compiled modules are not available
    show_compile_hint = True

    # whether when import a file the doc should be parsed - only activate when creating doc files
    parse_custom_docs = False

    @classmethod
    def data(cls):
        branch = "master"
        if 'rc' in __version__:
            branch = 'release'
        elif 'dev' in __version__:
            branch = 'develop'

        return f"https://raw.githubusercontent.com/anyoptimization/pymoo/{branch}/data/"


# returns the directory to be used for imports
def get_pymoo():
    return dirname(Config.root)



import numpy as np

from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.cython.stochastic_ranking import stochastic_ranking
from pymoo.core.population import Population


def find_ranks(l, s=None, start=1):
    assert len(l.shape) == 1, "Please provide a vector of shape one!"

    if s is None:
        s = l.argsort()

    ranks = np.zeros_like(s)

    rank = start
    val = l[s[0]]

    for j in s:
        if val != l[j]:
            rank += 1
        ranks[j] = rank

    return ranks


def cv_rank_sum(G):
    _, n_constr = G.shape

    # the ranks to be returned later on
    ranks = []

    # for each constraint value provided
    for g in G.T:

        # find the rank for each constraint - least infeasible solution will have rank 1
        r = find_ranks(np.maximum(0, g))

        # if there is at least one feasible solution - let us start from rank 0
        if np.any(g <= 0):
            r = r - 1

        ranks.append(r)

    ranks = np.column_stack(ranks)

    # now sum the ranks of all the constraints up
    rank_sum = ranks.sum(axis=1).astype(float)

    return rank_sum


def rel_eps_constr(a, b, eps):
    f_a, f_b, cv_a, cv_b = a.F[0], b.F[0], a.CV[0], b.CV[0]

    if (cv_a <= eps and cv_b <= eps) or cv_a == cv_b:
        if f_a < f_b:
            return 1.0
        elif f_a == f_b:
            return 0.0
        else:
            return -1.0
    else:
        if cv_a < cv_b:
            return 1.0
        else:
            return -1.0


def eps_replacement(pop, off, eps):
    repl = np.full(len(pop), False)
    for k in range(len(pop)):

        if rel_eps_constr(pop[k], off[k], eps[k]) <= 0:
            repl[k] = True

    return repl


class EpsilonConstraintReplacement(ReplacementSurvival):

    def __init__(self, eps=0.001, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps

    def _do(self, problem, pop, off, algorithm=None, **kwargs):
        n = len(pop)

        pop_X, pop_f, pop_cv = pop.get("X", "F", "CV")
        pop_f, pop_cv = pop_f[:, 0], pop_cv[:, 0]
        pop_feas = pop_cv <= 0

        infeas = np.sort(pop_cv[~pop_feas])
        n_infeas = len(pop) // 2

        if len(infeas) == 0:
            eps = np.inf
        else:
            k = min(n_infeas, len(infeas) - 1)
            eps = infeas[k]

        eps = eps * np.ones(n)

        return eps_replacement(pop, off, eps)


class FuzzyEpsilonConstraintReplacement(ReplacementSurvival):

    def __init__(self, eps=0.01, t=10, **kwargs):
        super().__init__(**kwargs)
        self.t = t
        self.eps = eps
        self.cnt = None

    def _do(self, problem, pop, off, algorithm=None, **kwargs):

        if self.cnt is None:
            self.cnt = np.zeros(len(pop), dtype=int)

        cnt = self.cnt

        cv = pop.get("CV")[:, 0]

        # cnt = self.cnt
        cnt = algorithm.n_gen - pop.get("n_gen") - 1

        # make sure we never replace the best solution if we would consider feasibility first
        best = FitnessSurvival().do(problem, Population.merge(pop, off), n_survive=1)[0]

        eps = np.zeros(len(pop))

        for k, t in enumerate(cnt):

            # cycle = (t // (4 * self.t))
            # max_eps = (2 ** cycle) * self.eps

            max_eps = self.eps

            t = t % (4 * self.t)

            if t < self.t:
                eps[k] = cv[k] + (max_eps - cv[k]) * (t / self.t)
            elif t < 2 * self.t:
                eps[k] = max_eps
            elif t < 3 * self.t:
                eps[k] = max_eps * (1 - ((t % self.t) / self.t))
            else:
                eps[k] = 0.0

        eps_is_zero = np.where(eps <= 0)[0]

        # print(len(eps_is_zero))

        repl = np.full(len(pop), False)
        for k in range(len(pop)):

            if pop[k] == best:
                repl[k] = False
            elif off[k] == best:
                repl[k] = True

            else:
                if rel_eps_constr(pop[k], off[k], eps[k]) <= 0:
                    repl[k] = True

        # self.cnt[repl] = 0
        # self.cnt[~repl] += 1

        return repl


class StochasticRankingImprovementReplacement(ReplacementSurvival):

    def _do(self, problem, pop, off, **kwargs):

        # merge the population for to find the ranks of them combined
        merged = Population.merge(pop, off)

        F, G = merged.get("F", "G")
        assert F.shape[1] == 1, "This survival only works for single-objective problems."
        f = F[:, 0]

        # for the constraint violation use the rank sum which makes the metric scale invariant
        cv = cv_rank_sum(G)
        # cv = merged.get("CV")[:, 0]

        merged.set("cv_rank_sum", cv)

        repl = np.full(len(pop), False)

        for k in range(len(pop)):
            a, b = pop[k], off[k]

            # cv_a, cv_b = a.get("cv_rank_sum"), b.get("cv_rank_sum")
            cv_a, cv_b = a.CV[0], b.CV[0]

            if np.random.random() < 0.25 or (cv_a == 0 and cv_b == 0):
                if b.F <= a.F:
                    repl[k] = True

            else:
                if cv_b <= cv_a:
                    repl[k] = True

        # make sure we never replace the best solution if we would consider feasibility first
        hierarch_sort = hierarchical_sort(f, cv)
        best = hierarch_sort[0]

        # if best from pop force not replacing it
        if best < len(pop):
            repl[best] = False

        # if best from offsprings force replacement
        else:
            repl[best - len(pop)] = True

        return repl

        # perform the stochastic sorting which gives infeasible solutions a chance to win over feasible once
        I = np.random.permutation(len(merged))
        S = I[stochastic_ranking(f, cv, 0.45, I=I)]

        # calculate the rank of each individual
        rank = S.argsort()

        rank_pop, rank_off = rank[:len(pop)], rank[len(pop):]
        repl = rank_off < rank_pop

        # make sure we never replace the best solution if we would consider feasibility first
        hierarch_sort = hierarchical_sort(f, cv)
        best = hierarch_sort[0]

        # if best from pop force not replacing it
        if best < len(pop):
            repl[best] = False

        # if best from offsprings force replacement
        else:
            repl[best - len(pop)] = True

        # set the rank to the all the individuals to be considered for later examples in an algorithm
        merged.set("sr", rank)

        return repl



import numpy as np

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival
from pymoo.docs import parse_doc_string
from pymoo.core.replacement import ImprovementReplacement
from pymoo.core.selection import Selection
from pymoo.operators.crossover.dex import DEX
from pymoo.operators.mutation.nom import NoMutation
from pymoo.operators.sampling.lhs import LHS
from pymoo.operators.selection.rnd import RandomSelection
from pymoo.util.display import SingleObjectiveDisplay
from pymoo.util.termination.default import SingleObjectiveDefaultTermination


# =========================================================================================================
# Implementation
# =========================================================================================================


class DE(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 n_offsprings=None,
                 sampling=LHS(),
                 variant="DE/best/1/bin",
                 CR=0.5,
                 F=None,
                 dither="vector",
                 jitter=False,
                 mutation=NoMutation(),
                 survival=None,
                 display=SingleObjectiveDisplay(),
                 **kwargs
                 ):
        """

        Parameters
        ----------

        pop_size : {pop_size}

        sampling : {sampling}

        variant : {{DE/(rand|best)/1/(bin/exp)}}
         The different variants of DE to be used. DE/x/y/z where x how to select individuals to be pertubed,
         y the number of difference vector to be used and z the crossover type. One of the most common variant
         is DE/rand/1/bin.

        F : float
         The F to be used during the crossover.

        CR : float
         The probability the individual exchanges variable values from the donor vector.

        dither : {{'no', 'scalar', 'vector'}}
         One strategy to introduce adaptive weights (F) during one run. The option allows
         the same dither to be used in one iteration ('scalar') or a different one for
         each individual ('vector).

        jitter : bool
         Another strategy for adaptive weights (F). Here, only a very small value is added or
         subtracted to the F used for the crossover for each individual.

        """

        # parse the information from the string
        _, sel, n_diff, mut, = variant.split("/")
        n_diffs = int(n_diff)
        if "-to-" in variant:
            n_diffs += 1

        selection = DES(sel)

        crossover = DEX(prob=1.0,
                        n_diffs=n_diffs,
                        F=F,
                        CR=CR,
                        variant=mut,
                        dither=dither,
                        jitter=jitter)

        super().__init__(pop_size=pop_size,
                         n_offsprings=n_offsprings,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         survival=survival,
                         display=display,
                         **kwargs)

        self.default_termination = SingleObjectiveDefaultTermination()

    def _initialize_advance(self, infills=None, **kwargs):
        self.pop = FitnessSurvival().do(self.problem, infills, n_survive=len(infills))

    def _infill(self):
        infills = self.mating.do(self.problem, self.pop, self.n_offsprings, algorithm=self)

        # tag each individual with an index - if a steady state version is executed
        infills.set("index", np.arange(len(infills)))

        # if number of offsprings is set lower than pop_size - randomly select
        if self.n_offsprings < self.pop_size:
            I = np.random.permutation(len(infills))[:self.n_offsprings]
            infills = infills[I]

        return infills

    def _advance(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus infills must to be provided."

        # get the indices where each offspring is originating from
        I = infills.get("index")

        # replace the individuals with the corresponding parents from the mating
        self.pop[I] = ImprovementReplacement().do(self.problem, self.pop[I], infills)

        # sort the population by fitness to make the selection simpler for mating (not an actual survival, just sorting)
        self.pop = FitnessSurvival().do(self.problem, self.pop)


# =========================================================================================================
# Selection and Mating
# =========================================================================================================

class DES(Selection):

    def __init__(self, variant) -> None:
        super().__init__()
        self.variant = variant

    def _do(self, pop, n_select, n_parents, **kwargs):
        variant = self.variant

        # create offsprings and add it to the data of the algorithm
        P = RandomSelection().do(pop, n_select, n_parents)

        if variant == "best":
            P[:, 0] = 0
        elif variant == "current-to-best":
            P[:, 0] = np.arange(len(pop))
            P[:, 1] = 0
            P[:, 2] = np.arange(len(pop))
        elif variant == "current-to-rand":
            P[:, 0] = np.arange(len(pop))
            P[:, 2] = np.arange(len(pop))
        elif variant == "rand-to-best":
            P[:, 1] = 0
            P[:, 2] = np.arange(len(pop))
        elif variant == "current-to-pbest":
            # best 10% of the population
            n_pbest = int(np.ceil(0.1 * len(pop)))

            # the corresponding indices to select from
            pbest = np.arange(n_pbest)

            P[:, 0] = np.arange(len(pop))
            P[:, 1] = np.random.choice(pbest, len(pop))
            P[:, 2] = np.arange(len(pop))

        return P


parse_doc_string(DE.__init__)



import numpy as np

from pymoo.util.ref_dirs.optimizer import GradientDescent


class CoordinateDescent(GradientDescent):

    def direction(self, dF, **kwargs):
        dF = dF[0]
        p = np.zeros(len(dF))
        k = np.abs(dF).argmax()
        p[k] = 1 * np.sign(dF[k])
        return - p



import matplotlib.pyplot as plt
import numpy as np
import scipy.stats
from prettytable import PrettyTable

from pymoo.algorithms.moo.gde3 import GDE3
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.algorithms.soo.nonconvex.cmaes import SimpleCMAES
from pymoo.algorithms.soo.nonconvex.de import DE
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.algorithms.soo.nonconvex.nelder_mead import NelderMead
from pymoo.algorithms.soo.nonconvex.pattern_search import PatternSearch
from pymoo.algorithms.soo.nonconvex.pso import PSO
from pymoo.experimental.benchmarking.analyzer.convergence import ConvergenceAnalyzer
from pymoo.experimental.benchmarking.analyzer.groupby import GroupBy
from pymoo.experimental.benchmarking.analyzer.moo import MultiObjectiveAnalyzer
from pymoo.experimental.benchmarking.analyzer.others import MultiObjectiveConvergenceAnalyzer
from pymoo.experimental.benchmarking.analyzer.soo import SingleObjectiveAnalyzer
from pymoo.experimental.benchmarking.benchmark import Benchmark, DefaultWriter, DefaultLoader
from pymoo.experimental.benchmarking.recoder import DefaultSingleObjectiveRecorder, DefaultMultiObjectiveRecorder
from pymoo.experimental.benchmarking.util import filter_by
from pymoo.factory import get_problem, ZDT1, ZDT2, ZDT3
from pymoo.util.termination.max_eval import MaximumFunctionCallTermination

if __name__ == "__main__":
    FOLDER = "/Users/blankjul/moo_benchmark"

    recorder = None
    recorder = DefaultMultiObjectiveRecorder()

    benchmark = Benchmark(n_runs=11, recorder=recorder)

    benchmark.add_problem("zdt1", ZDT1(), termination=("n_gen", 200))
    # benchmark.add_problem("zdt2", ZDT2(), termination=("n_gen", 200))
    # benchmark.add_problem("zdt3", ZDT3(), termination=("n_gen", 200))

    benchmark.add_algorithm("nsga2", NSGA2())
    benchmark.add_algorithm("gde3", GDE3())

    loader = DefaultLoader(FOLDER)
    writer = DefaultWriter(FOLDER)

    results = benchmark.run(writer=writer,
                            loader=loader,
                            run_if_loading_fails=True)

    # set the igd values for each of the problems
    MultiObjectiveAnalyzer().run(results, benchmark=benchmark, inplace=True)

    # now aggregate all the runs to have some representative values
    attrs = [("igd", np.array, "igd"),
             ("igd", np.mean, "avg"),
             ("igd", np.std, "std")]

    igd = GroupBy(attrs).run(results, group_by=["problem", "algorithm"])

    for scope, d in filter_by(igd, ["problem"], return_group=True):

        # find the best algorithm for this problem
        l = sorted(d, key=lambda x: x["avg"])
        best = l[0]["igd"]

        t = PrettyTable()
        t.title = scope["problem"]
        t.field_names = ['Algorithm', 'avg', 'std', 'shapiro', 'levene', 't-test', 'wilcoxon']

        for i, e in enumerate(l):
            f = e["igd"]

            _, pval = scipy.stats.shapiro(f)
            shapiro = "*" if pval >= 0.01 else ""

            _, pval = scipy.stats.levene(best, f)
            levene = "* (%.3f)" % pval if pval >= 0.05 else ""

            _, pval = scipy.stats.ttest_ind(f, best, alternative="greater")
            ttest = "* (%.3f)" % pval if pval >= 0.05 else ""

            if len(best) == len(f):
                _, pval = scipy.stats.wilcoxon(f, best, zero_method="zsplit", alternative="greater")
                wilcoxon = "* (%.3f)" % pval if pval >= 0.05 else ""
            else:
                wilcoxon = "x"

            t.add_row([e["algorithm"], "%.10f" % e["avg"], "%.10f" % e["std"], shapiro, levene, ttest, wilcoxon])

        print(t)
        print()





import numpy as np
from numpy.random import permutation

from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.docs import parse_doc_string
from pymoo.core.population import Population
from pymoo.core.selection import Selection
from pymoo.operators.crossover.dex import DEX
from pymoo.operators.mutation.pm import PM
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.selection.rnd import RandomSelection
from pymoo.util.display import MultiObjectiveDisplay
from pymoo.util.normalization import ObjectiveSpaceNormalization
from pymoo.util.optimum import filter_optimum
from pymoo.util.termination.max_eval import MaximumFunctionCallTermination
from pymoo.util.termination.max_gen import MaximumGenerationTermination


def norm(V):
    return V / np.sqrt(np.sum(V ** 2, axis=1))[:, None]


class NeighborhoodSelection(Selection):

    def __init__(self, n_neighbors=10) -> None:
        super().__init__()
        self.n_neighbors = n_neighbors

    def _do(self, pop, n_select, n_parents, ideal=None, **kwargs):
        assert n_parents == 3, "This selection is based on three parents!"
        assert ideal is not None, "The neighborhood selection needs the ideal point for normalization!"

        F = pop.get("F")
        F = F - ideal + 1e-64
        F = norm(F)

        cosv = F @ F.T - 3 * np.eye(len(F))
        I = np.argsort(-cosv, axis=1)
        neighbours = I[:, :self.n_neighbors]

        P = RandomSelection().do(pop, n_select, n_parents)
        P[:, 0] = np.arange(len(P))

        for k in range(len(pop)):

            if np.random.random() < 0.7:
                P[k, 1:] = np.random.choice(neighbours[k], 2, replace=False)
            else:
                P[k, 1:2] = np.random.choice(neighbours[k], 1, replace=False)

        return P


def Indicator_based_CHT(F, indicator, W, N):
    IrFitness = indicator.min(axis=1)
    dom = np.where(IrFitness >= 0)[0]

    if len(dom) <= N:
        return np.argsort(-IrFitness)[:N]
    else:

        F = F[dom]
        indicator = np.copy(indicator[dom][:, dom])

        survivors = np.full(len(F), True)

        # normalize the weights and the objectives
        nW, nF = norm(W), norm(F)

        # assigned each solution based on the angle to a niche
        A = np.argmax((nF @ nW.T), axis=1)

        # create an assignment list for each niche
        niches = [[] for _ in range(len(W))]
        [niches[j].append(i) for i, j in enumerate(A)]

        # the niche counts
        cnt = [len(inds) for inds in niches]

        neighbor = indicator.argmin(axis=1)
        vals = indicator.min(axis=1)

        while survivors.sum() > N:

            # find the niche where a solution should be deleted
            niche = np.argmax(cnt)

            # all solutions in that niche
            I = niches[niche]

            # select the solution in this niche with the worst indicator value
            j = I[vals[I].argmin()]

            # set the values in the indicator matrix to infinity
            indicator[j, :] = np.inf
            indicator[:, j] = np.inf
            vals[j] = np.inf

            # now update the neighbors and matrix partially (only the one that have changed)
            K = np.where(j == neighbor)[0]
            if len(K) > 0:
                neighbor[K] = indicator[K].argmin(axis=1)
                vals[K] = indicator[K].min(axis=1)

            # remove the solution from the niche
            niches[niche] = [e for e in I if e != j]
            cnt[niche] -= 1

            # deselect the solution to survive
            survivors[j] = False

        return dom[survivors]


def Selection_Operator_of_PREA(F, indicator, N):
    fitness = indicator.min(axis=1)
    J = np.where(fitness >= 0)[0]

    if len(J) <= N:
        return np.argsort(-fitness)[:N]
    else:

        survivors = np.full(len(J), True)

        _F = F[J]

        ir = np.copy(indicator[J][:, J])
        ir_neighbor = ir.argmin(axis=1)
        ir_vals = ir.min(axis=1)

        while survivors.sum() > N:

            # select the solution in this niche with the worst indicator value
            j = ir_vals.argmin()

            # set the values in the indicator matrix to infinity
            ir[j] = np.inf
            ir[:, j] = np.inf

            # now update the neighbors and matrix partially (only the one that have changed)
            K = np.where(j == ir_neighbor)[0]
            if len(K) > 0:
                ir_neighbor[K] = ir[K].argmin(axis=1)
                ir_vals[K] = ir[K].min(axis=1)

            # the value of the solution itself becomes infinity
            ir_vals[j] = np.inf

            # deselect the solution to survive
            survivors[j] = False

        nadir = _F[survivors].max(axis=0)

        J = np.where(np.all(F <= nadir, axis=1))[0]

        if len(J) <= N:
            return J

        _F = F[J] / nadir

        survivors = np.full(len(_F), True)

        ir = np.copy(indicator[J][:, J])
        ir_neighbor = ir.argmin(axis=1)
        ir_vals = ir.min(axis=1)

        dist = np.zeros((len(_F), len(_F)))
        for i in range(len(_F)):
            Fi = _F[i]
            Fdelta = _F - Fi
            dist[i, :] = np.sqrt(np.sum(Fdelta ** 2, axis=1) - np.sum(Fdelta, axis=1) ** 2 / _F.shape[1])
            dist[i, i] = np.inf

        dist_neighbor = np.argmin(dist, axis=1)
        dist_vals = np.min(dist, axis=1)

        while survivors.sum() > N:

            a = np.argmin(dist_vals)
            b = dist_neighbor[a]

            if ir_vals[a] < ir_vals[b]:
                j = a
            else:
                j = b

            # set the values in the distance matrix to infinity
            dist[j, :] = np.inf
            dist[:, j] = np.inf

            # now update the neighbors and matrix partially (only the one that have changed)
            K = np.where(j == dist_neighbor)[0]
            if len(K) > 0:
                dist_neighbor[K] = dist[K].argmin(axis=1)
                dist_vals[K] = dist[K].min(axis=1)

            # the value of the solution itself becomes infinity
            dist_vals[j] = np.inf

            # set the values in the indicator matrix to infinity
            ir[j] = np.inf
            ir[:, j] = np.inf

            # now update the neighbors and matrix partially (only the one that have changed)
            K = np.where(j == ir_neighbor)[0]
            if len(K) > 0:
                ir_neighbor[K] = ir[K].argmin(axis=1)
                ir_vals[K] = ir[K].min(axis=1)

            # the value of the solution itself becomes infinity
            ir_vals[j] = np.inf

            # deselect the solution to survive
            survivors[j] = False

        return J[survivors]


class ICMA(GeneticAlgorithm):

    def __init__(self,
                 ref_dirs,
                 pop_size=None,
                 sampling=FloatRandomSampling(),
                 selection=NeighborhoodSelection(),
                 crossover=DEX(prob=0.9, CR=0.5, variant='bin'),
                 mutation=PM(eta=20),
                 display=MultiObjectiveDisplay(),
                 **kwargs):

        # set reference directions and pop_size
        self.ref_dirs = ref_dirs
        if self.ref_dirs is not None:
            if pop_size is None:
                pop_size = len(self.ref_dirs)

        super().__init__(pop_size=pop_size,
                         sampling=sampling,
                         selection=selection,
                         crossover=crossover,
                         mutation=mutation,
                         eliminate_duplicates=False,
                         display=display,
                         **kwargs)

        self.RA = None
        self.norm = None
        self.archive = None

    def _setup(self, problem, **kwargs):

        self.norm = ObjectiveSpaceNormalization()

        # if maximum functions termination convert it to generations
        if isinstance(self.termination, MaximumFunctionCallTermination):
            n_gen = np.ceil((self.termination.n_max_evals - self.pop_size) / self.n_offsprings)
            self.termination = MaximumGenerationTermination(n_gen)

        # check whether the n_gen termination is used - otherwise this algorithm can be not run
        if not isinstance(self.termination, MaximumGenerationTermination):
            raise Exception("Please use the n_gen or n_eval as a termination criterion to run ICMA!")

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills, **kwargs)
        self.norm.update(infills)
        self.archive = infills
        self.Ra = 1.0

    def _infill(self, **kwargs):

        pop, archive, Ra = self.pop, self.archive, self.Ra

        N = self.pop_size
        Nt = int(np.floor(Ra * N))

        from_pop = pop[permutation(len(pop))[:Nt]]
        from_archive = pop[permutation(len(archive))[:N - Nt]]
        pool = Population.merge(from_pop, from_archive)

        # first do the differential evolution mating
        off = self.mating.do(self.problem, pool, self.n_offsprings, algorithm=self,
                             ideal=self.norm.ideal(only_feas=False))

        return off

    def _advance(self, infills=None, **kwargs):
        self.norm.update(infills)

        pop = Population.merge(Population.merge(self.pop, infills), self.archive)

        F = pop.get("F")

        if self.problem.has_constraints():
            G = pop.get("G")
            C = np.maximum(0, G)
            C = C / np.maximum(1, C.max(axis=0))
            CV = C.sum(axis=1)
        else:
            CV = np.zeros(len(pop))

        F = F - self.norm.ideal(only_feas=False) + 1e-6

        indicator = np.full((len(pop), len(pop)), np.inf)

        for i in range(len(pop)):
            Ci = CV[i]
            Fi = F[i]

            if Ci == 0:

                Ir = np.log(Fi / F)
                MaxIr = np.max(Ir, axis=1)
                MinIr = np.min(Ir, axis=1)

                CVA = MaxIr

                J = MaxIr <= 0
                CVA[J] = MinIr[J]

                val = CVA

            else:
                IC = (Ci + 1e-6) / (CV + 1e-6)
                CVF = np.max(np.maximum(Fi, F) / np.minimum(Fi, F), axis=1)
                val = np.log(np.maximum(CVF, IC))

            indicator[:, i] = val
            indicator[i, i] = np.inf

        feas = np.where(CV <= 0)[0]

        if len(feas) <= self.pop_size:
            self.archive = pop[np.argsort(CV)[:self.pop_size]]
        else:
            feas_F = pop[feas].get("F") - self.norm.ideal(only_feas=True) + 1e-6
            I = Selection_Operator_of_PREA(feas_F, indicator[feas][:, feas], self.pop_size)
            self.archive = pop[feas[I]]

        I = Indicator_based_CHT(F, indicator, self.ref_dirs, self.pop_size)
        self.pop = pop[I]

        self.Ra = 1 - self.n_gen / self.termination.n_max_gen

    def _set_optimum(self, **kwargs):
        self.opt = filter_optimum(self.archive, least_infeasible=True)


parse_doc_string(ICMA.__init__)



import numpy as np

from pymoo.experimental.deriv import DerivationBasedAlgorithm
from pymoo.algorithms.base.line import LineSearchProblem
from pymoo.algorithms.soo.univariate.exp import ExponentialSearch
from pymoo.algorithms.soo.univariate.golden import GoldenSectionSearch
from pymoo.core.population import Population
from pymoo.util.vectors import max_alpha


class GradientDescent(DerivationBasedAlgorithm):

    def direction(self, dF, **kwargs):
        return - dF

    def step(self):
        problem, sol = self.problem, self.opt[0]
        self.evaluator.eval(self.problem, sol, evaluate_values_of=["dF"])
        dF = sol.get("dF")[0]

        print(sol)

        if np.linalg.norm(dF) ** 2 < 1e-8:
            self.termination.force_termination = True
            return

        direction = self.direction(dF)

        line = LineSearchProblem(self.problem, sol, direction, strict_bounds=self.strict_bounds)
        alpha = self.alpha

        if self.strict_bounds:

            if problem.has_bounds():
                line.xu = np.array([max_alpha(sol.X, direction, *problem.bounds(), mode="all_hit_bounds")])

            # remember the step length from the last run
            alpha = min(alpha, line.xu[0])

            if alpha == 0:
                self.termination.force_termination = True
                return

        # make the solution to be the starting point of the univariate search
        x0 = sol.copy(deep=True)
        x0.set("__X__", x0.get("X"))
        x0.set("X", np.zeros(1))

        # determine the brackets to be searched in
        exp = ExponentialSearch(delta=alpha).setup(line, evaluator=self.evaluator, termination=("n_iter", 20), x0=x0)
        a, b = exp.run().pop[-2:]

        # search in the brackets
        res = GoldenSectionSearch().setup(line, evaluator=self.evaluator, termination=("n_iter", 20), a=a, b=b).run()
        infill = res.opt[0]

        # set the alpha value and revert the X to be the multi-variate one
        infill.set("X", infill.get("__X__"))
        self.alpha = infill.get("alpha")[0]

        # keep always a few historical solutions
        self.pop = Population.merge(self.pop, infill)[-10:]



from pymoo.core.algorithm import Algorithm


class MetaAlgorithm(Algorithm):

    def __init__(self,
                 clazz,
                 *args,
                 display=None,
                 **kwargs
                 ):
        super().__init__(*args, display=display, **kwargs)

        self.clazz = clazz
        self.args = args
        self.kwargs = kwargs
        self.algorithm = self.create()

    def create(self, *args, **kwargs):
        return self.clazz(*self.args, *args, **self.kwargs, **kwargs)

    def setup(self, *args, **kwargs):
        self.algorithm.setup(*args, **kwargs)
        super().setup(*args, **kwargs)

    def _initialize_infill(self):
        return self._infill()

    def _initialize_advance(self, **kwargs):
        return self._advance(**kwargs)

    def _infill(self):
        return self.algorithm.infill()

    def _advance(self, infills=None, **kwargs):
        return self.algorithm.advance(infills=infills, **kwargs)

    def _set_optimum(self):
        self.opt = self.algorithm.opt






import numpy as np

from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.core.problem import Problem
from pymoo.util.misc import at_least_2d_array


def set_feasibility(pop):
    for ind in pop:
        cv = ind.get("CV")
        if cv is not None:
            ind.feasible = cv <= 0


def set_cv(pop, feasbility=True):
    for ind in pop:
        if ind.G is None:
            ind.CV = np.zeros(1)
        else:
            ind.CV = Problem.calc_constraint_violation(at_least_2d_array(ind.G))[0]

    if feasbility:
        set_feasibility(pop)


class Evaluator:
    """

    The evaluator class which is used during the algorithm execution to limit the number of evaluations.
    This can be based on convergence, maximum number of evaluations, or other criteria.

    """

    def __init__(self,
                 skip_already_evaluated=True,
                 evaluate_values_of=["F", "CV", "G"]):
        self.n_eval = 0
        self.evaluate_values_of = evaluate_values_of
        self.skip_already_evaluated = skip_already_evaluated

    def eval(self,
             problem,
             pop,
             skip_already_evaluated=None,
             evaluate_values_of=None,
             count_evals=True,
             **kwargs):
        """

        This function is used to return the result of one valid evaluation.

        Parameters
        ----------
        problem : class
            The problem which is used to be evaluated
        pop : np.array or Population object
        kwargs : dict
            Additional arguments which might be necessary for the problem to evaluate.

        """

        if evaluate_values_of is None:
            evaluate_values_of = ["F", "G", "CV", "feasible"]

        is_individual = isinstance(pop, Individual)
        is_numpy_array = isinstance(pop, np.ndarray) and not isinstance(pop, Population)

        # make sure the object is a population
        if is_individual or is_numpy_array:
            pop = Population().create(pop)

        # find indices to be evaluated
        if skip_already_evaluated or (skip_already_evaluated is None and self.skip_already_evaluated):
            I = []
            for i, individual in enumerate(pop):

                # if the evaluated has been overwritten to be none or not a set
                if not isinstance(individual.evaluated, set):
                    individual.evaluated = set()

                # if not everything has been evaluated
                if not all([e in individual.evaluated for e in evaluate_values_of]):
                    I.append(i)

        else:
            I = np.arange(len(pop))

        # update the function evaluation counter
        if count_evals:
            self.n_eval += len(I)

        # actually evaluate all solutions using the function that can be overwritten
        if len(I) > 0:
            self._eval(problem, pop[I], evaluate_values_of=evaluate_values_of, **kwargs)

            # set the feasibility attribute if cv exists
            set_feasibility(pop[I])

        if is_individual:
            return pop[0]
        elif is_numpy_array:
            if len(pop) == 1:
                pop = pop[0]
            return tuple([pop.get(e) for e in self.evaluate_values_of])
        else:
            return pop

    def _eval(self, problem, pop, evaluate_values_of=None, **kwargs):
        evaluate_values_of = self.evaluate_values_of if evaluate_values_of is None else evaluate_values_of

        out = problem.evaluate(pop.get("X"),
                               return_values_of=evaluate_values_of,
                               return_as_dictionary=True,
                               **kwargs)

        for key, val in out.items():
            if val is None:
                continue
            else:
                pop.set(key, val)

        for ind in pop:
            ind.evaluated.update(out.keys())


class VoidEvaluator(Evaluator):

    def __init__(self, value=np.inf, **kwargs):
        super().__init__(**kwargs)
        self.value = value

    def eval(self, problem, pop, **kwargs):
        val = self.value
        if val is not None:
            for individual in pop:
                if individual.F is None:
                    individual.F = np.full(problem.n_obj, val)
                    individual.G = np.full(problem.n_constr, val) if problem.has_constraints() else None
                    individual.CV = [-np.inf]
                    individual.feasible = [False]



import os
import warnings

import numpy as np

from pymoo.core.indicator import Indicator
from pymoo.indicators.distance_indicator import derive_ideal_and_nadir_from_pf
from pymoo.util.misc import at_least_2d_array
from pymoo.util.nds.non_dominated_sorting import NonDominatedSorting
from pymoo.vendor.hv import HyperVolume as _HyperVolume


def hypervolume_by_command(path_to_hv, X, ref_point):
    """
    A method to manually call the Hypervolume calculation if it is installed.
    http://lopez-ibanez.eu/hypervolume


    Parameters
    ----------
    path_to_hv : Path to the compiled executable
    X : Points to calculate the Hypervolume
    ref_point : Reference Point

    """

    ref_point_as_str = " ".join(format(x, ".3f") for x in ref_point)

    current_folder = os.path.dirname(os.path.abspath(__file__))

    path_to_input = os.path.join(current_folder, "in.dat")
    np.savetxt(path_to_input, X)

    path_to_output = os.path.join(current_folder, "out.dat")

    command = "%s -r \"%s\" %s > %s" % (path_to_hv, ref_point_as_str, path_to_input, path_to_output)
    # print(command)
    os.system(command)

    with open(path_to_output, 'r') as f:
        val = f.read()

    os.remove(path_to_input)
    os.remove(path_to_output)

    try:
        hv = float(val)
    except:
        warnings.warn(val)
        return - np.inf

    return hv


class Hypervolume(Indicator):

    def __init__(self, ref_point=None, pf=None, nds=True, norm_ref_point=True, ideal=None, nadir=None, **kwargs):
        pf = at_least_2d_array(pf, extend_as="row")
        ideal, nadir = derive_ideal_and_nadir_from_pf(pf, ideal=ideal, nadir=nadir)

        super().__init__(ideal=ideal, nadir=nadir, **kwargs)

        # whether the input should be checked for domination or not
        self.nds = nds

        # the reference point that shall be used - either derived from pf or provided
        ref_point = ref_point
        if ref_point is None:
            if pf is not None:
                ref_point = pf.max(axis=0)

        # we also have to normalize the reference point to have the same scales
        if norm_ref_point:
            ref_point = self.normalization.forward(ref_point)

        self.ref_point = ref_point
        assert self.ref_point is not None, "For Hypervolume a reference point needs to be provided!"

    def _do(self, F):
        if self.nds:
            non_dom = NonDominatedSorting().do(F, only_non_dominated_front=True)
            F = np.copy(F[non_dom, :])

        # calculate the hypervolume using a vendor library
        hv = _HyperVolume(self.ref_point)
        val = hv.compute(F)
        return val



import numpy as np

from pymoo.algorithms.soo.nonconvex.es import ES
from pymoo.docs import parse_doc_string
from pymoo.core.survival import Survival
from pymoo.util.function_loader import load_function


class StochasticRankingSurvival(Survival):

    def __init__(self, PR):
        super().__init__(filter_infeasible=False)
        self.PR = PR

    def _do(self, problem, pop, *args, n_survive=None, **kwargs):
        assert problem.n_obj == 1, "This stochastic ranking implementation only works for single-objective problems."

        F, G = pop.get("F", "G")
        f = F[:, 0]

        if problem.n_constr == 0:
            I = f.argsort()

        else:
            phi = (np.maximum(0, G) ** 2).sum(axis=1)
            J = np.arange(len(phi))
            I = load_function("stochastic_ranking")(f, phi, self.PR, J)

        return pop[I][:n_survive]


class SRES(ES):

    def __init__(self, PF=0.45, **kwargs):
        """
        Stochastic Ranking Evolutionary Strategy (SRES)

        Parameters
        ----------
        PF: float
            The stochastic ranking weight for choosing a random decision while doing the modified bubble sort.
        """
        super().__init__(survival=StochasticRankingSurvival(PF), **kwargs)
        self.PF = PF


parse_doc_string(SRES.__init__)



import matplotlib.pyplot as plt
import numpy as np

from pymoo.util.termination.f_tol import MultiObjectiveSpaceToleranceTerminationWithRenormalization
from pymoo.visualization.video.callback_video import AnimationCallback


class RunningMetric(AnimationCallback):

    def __init__(self,
                 delta_gen,
                 n_plots=4,
                 only_if_n_plots=False,
                 key_press=True,
                 **kwargs) -> None:

        super().__init__(**kwargs)
        self.delta_gen = delta_gen
        self.key_press = key_press
        self.only_if_n_plots = only_if_n_plots
        self.term = MultiObjectiveSpaceToleranceTerminationWithRenormalization(n_last=100000,
                                                                               all_to_current=True,
                                                                               sliding_window=False)
        self.hist = []
        self.n_plots = n_plots

    def do(self, _, algorithm, force_plot=False, **kwargs):
        self.term.do_continue(algorithm)

        metric = self.term.get_metric()
        metrics = self.term.metrics
        tau = len(metrics)

        fig = None

        if metric is not None and (tau + 1) % self.delta_gen == 0 or force_plot:

            _delta_f = metric["delta_f"]
            _delta_ideal = [m['delta_ideal'] > 0.005 for m in metrics]
            _delta_nadir = [m['delta_nadir'] > 0.005 for m in metrics]

            if force_plot or not self.only_if_n_plots or (self.only_if_n_plots and len(self.hist) == self.n_plots - 1):

                fig, ax = plt.subplots()

                if self.key_press:
                    def press(event):
                        if event.key == 'q':
                            algorithm.termination.force_termination = True

                    fig.canvas.mpl_connect('key_press_event', press)

                for k, f in self.hist:
                    ax.plot(np.arange(len(f)) + 1, f, label="t=%s" % (k+1), alpha=0.6, linewidth=3)
                ax.plot(np.arange(len(_delta_f)) + 1, _delta_f, label="t=%s (*)" % (tau+1), alpha=0.9, linewidth=3)

                for k in range(len(_delta_ideal)):
                    if _delta_ideal[k] or _delta_nadir[k]:
                        ax.plot([k+1, k+1], [0, _delta_f[k]], color="black", linewidth=0.5, alpha=0.5)
                        ax.plot([k+1], [_delta_f[k]], "o", color="black", alpha=0.5, markersize=2)

                ax.set_yscale("symlog")
                ax.legend()

                ax.set_xlabel("Generation")
                ax.set_ylabel("$\Delta \, f$", rotation=0)

                if self.key_press:
                    plt.draw()
                    plt.waitforbuttonpress()
                    plt.close('all')

            self.hist.append((tau, _delta_f))
            self.hist = self.hist[-(self.n_plots - 1):]





import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.operators.repair.inverse_penalty import inverse_penality


def eps_if_less_than_eps(x, eps):
    x[x < eps] = eps
    return x


def pcx(X, k, eta, zeta):
    # the number of parents to be considered
    n_parents, n_var = X.shape

    # find the centroid of the parents
    centroid = np.mean(X, axis=0)

    # calculate the difference between the centroid and the k-th parent
    diff_to_centroid = centroid - X[k]
    dist_to_centroid = np.linalg.norm(diff_to_centroid)

    # calculate the differences from all parents to parent k
    diff_to_index = X - X[k]
    dist_to_index = np.linalg.norm(diff_to_index, axis=1)

    # orthogonal directions are computed
    orth_dir = np.zeros(n_parents)

    S = dist_to_index > 1e-16
    if S.sum() == 0:
        return centroid

    for i in range(n_parents):
        if S[i]:
            temp1 = (diff_to_index[i] * diff_to_centroid).sum()
            temp2 = temp1 / (dist_to_index[i] * dist_to_centroid)
            temp3 = max(1.0 - temp2 ** 2, 0)
            orth_dir[i] = dist_to_index[i] * temp3 ** 0.5

    # this is the avg of the perpendicular distances from other parents to the parent k
    D_not = orth_dir.sum() / (n_parents - 1)

    # generating zero-mean normally distributed variables
    mu, sigma = 0.0, (D_not * eta)
    rnd = np.random.normal(mu, sigma, n_var)

    # implemented just like the c code - generate_new.h file
    noise = rnd - (np.sum(rnd * diff_to_centroid) * diff_to_centroid) / dist_to_centroid ** 2

    bias_to_centroid = diff_to_centroid * np.random.normal(0.0, zeta, 1)

    off = X[k] + noise + bias_to_centroid

    return off


class ParentCentricCrossover(Crossover):
    def __init__(self,
                 eta=0.1,
                 zeta=0.1,
                 **kwargs):

        super().__init__(n_parents=3, n_offsprings=1, **kwargs)
        self.eta = float(eta)
        self.zeta = float(zeta)

    def _do(self, problem, X, **kwargs):
        n_parents, n_matings, n_var = X.shape

        off = np.empty([self.n_offsprings, n_matings, n_var])
        K = np.row_stack([np.random.permutation(self.n_parents) for _ in range(n_matings)])[:, :self.n_offsprings]

        for j in range(self.n_offsprings):
            for i in range(n_matings):
                x = X[:, i, :]
                k = K[i, j]

                # do the crossover
                _off = pcx(x, k, self.eta, self.zeta)

                # make sure the offspring is in bounds and assign
                off[j, i, :] = inverse_penality(_off, x[k], problem.xl, problem.xu)

        return off


class PCX(ParentCentricCrossover):
    pass



from pymoo.algorithms.soo.convex.deriv.gd import GradientDescent
from pymoo.factory import Sphere
from pymoo.operators.sampling.random_sampling import random
from pymoo.optimize import minimize
from pymoo.problems.autodiff import AutomaticDifferentiation


problem = AutomaticDifferentiation(Sphere())
problem.xl += 10
problem.xu += 11

X = random(problem, 1)

algorithm = GradientDescent(strict_bounds=False)

res = minimize(problem,
               algorithm,
               seed=1,
               verbose=True)

print(res.X)
print(res.F)



from copy import deepcopy

import numpy as np

from pymoo.core.callback import Callback
from pymoo.indicators.igd import IGD


class Recorder(Callback):

    def __init__(self) -> None:
        super().__init__()
        self.data = []

    def notify(self, algorithm, **kwargs):
        elem = self.save(algorithm)
        self.data.append(elem)

    def save(self, algorithm):
        pass


class DefaultSingleObjectiveRecorder(Recorder):

    def save(self, algorithm):
        opt = deepcopy(algorithm.opt[0])
        n_evals, cv, f = algorithm.evaluator.n_eval, opt.CV[0], opt.F[0]

        # see if the problem has the best value stored
        pf = algorithm.problem.pareto_front()
        fgap = None

        # if we have on let us measure how close we are
        if pf is not None:
            fopt = pf.min()
            fgap = f - fopt

        return dict(n_evals=n_evals, cv=float(cv), f=float(f), fgap=fgap)


class DefaultMultiObjectiveRecorder(Recorder):

    def save(self, algorithm):

        min_cv = algorithm.opt.get("CV").min()

        igd = None

        pf = algorithm.problem.pareto_front()
        if pf is not None:
            if min_cv <= 0:
                igd = IGD(pf=pf, zero_to_one=True).do(algorithm.opt.get("F"))
            else:
                igd = np.inf

        return {
            "n_evals": algorithm.evaluator.n_eval,
            "cv": min_cv,
            "igd": igd,
        }



import numpy as np

from pymoo.experimental.benchmarking.analyzer.analyzer import Analyzer
from pymoo.indicators.igd import IGD
from pymoo.util.misc import from_dict


class MultiObjectiveAnalyzer(Analyzer):

    def do(self, data, scope=None, benchmark=None, inplace=False, **kwargs):
        assert benchmark is not None, "The benchmark is necessary to retrieve the known optimum of a function"

        problem = benchmark.problems[data["problem"]]["obj"]
        CV, F = from_dict(data, "CV", "F")

        igd = np.inf
        pf = problem.pareto_front(**kwargs)
        if pf is not None:
            igd = IGD(pf, zero_to_one=True).do(F)

        ret = {
            "pf": pf,
            "igd": igd,
        }

        if inplace:
            for k, v in ret.items():
                data[k] = v

        return ret







import matplotlib.pyplot as plt
import numpy as np
import scipy.stats
from prettytable import PrettyTable

from pymoo.algorithms.soo.nonconvex.cmaes import SimpleCMAES
from pymoo.algorithms.soo.nonconvex.de import DE
from pymoo.algorithms.soo.nonconvex.ga import GA
from pymoo.algorithms.soo.nonconvex.nelder_mead import NelderMead
from pymoo.algorithms.soo.nonconvex.pattern_search import PatternSearch
from pymoo.algorithms.soo.nonconvex.pso import PSO
from pymoo.experimental.benchmarking.analyzer.convergence import ConvergenceAnalyzer
from pymoo.experimental.benchmarking.analyzer.groupby import GroupBy
from pymoo.experimental.benchmarking.analyzer.soo import SingleObjectiveAnalyzer
from pymoo.experimental.benchmarking.benchmark import Benchmark, DefaultWriter, DefaultLoader
from pymoo.experimental.benchmarking.recoder import DefaultSingleObjectiveRecorder
from pymoo.experimental.benchmarking.util import filter_by
from pymoo.factory import get_problem
from pymoo.util.termination.max_eval import MaximumFunctionCallTermination

if __name__ == "__main__":

    FOLDER = "/Users/blankjul/my_benchmark"

    recorder = DefaultSingleObjectiveRecorder()

    termination = MaximumFunctionCallTermination(100 * 1000)

    benchmark = Benchmark(n_runs=11, recorder=recorder)

    # benchmark.add_problem("sphere-10d", Sphere(n_var=10), ("n_evals", 1000))
    # benchmark.add_problem("himmelblau", Himmelblau(), ("n_evals", 400))
    # benchmark.add_problem("rosenbrock", Rosenbrock(), ("n_evals", 500))
    # benchmark.add_problem("ackley-10d", Ackley(n_var=10), ("n_evals", 1000))
    # benchmark.add_problem("rastrigin-5d", Rastrigin(n_var=5), ("n_evals", 1000))

    instance = 1
    n_evals = 1000
    for n_var in [10, 20, 40]:
        for function in range(1, 25):
            label = f"bbob-f{function:02d}-{instance}"
            benchmark.add_problem(label + "-" + str(n_var), get_problem(label, n_var=n_var), ("n_evals", n_evals))

    benchmark.add_algorithm("de", DE())
    benchmark.add_algorithm("ga", GA())
    benchmark.add_algorithm("pso", PSO())
    benchmark.add_algorithm("cmaes", SimpleCMAES())
    benchmark.add_algorithm("ps", PatternSearch())
    benchmark.add_algorithm("nm", NelderMead())

    loader = DefaultLoader(FOLDER)
    # loader = None

    writer = DefaultWriter(FOLDER)

    results = benchmark.run(writer=writer,
                            loader=loader,
                            run_if_loading_fails=True)

    # _ = SingleObjectiveAnalyzer().run(results, benchmark=benchmark, inplace=True)
    results = SingleObjectiveAnalyzer().run(results, benchmark=benchmark, inplace=False)

    attrs = {"fgap": ["np.median", "np.mean", "np.std", "collect"]}
    mean = GroupBy(attrs).run(results, group_by=["problem", "algorithm"])

    for scope, d in filter_by(mean, ["problem"], return_group=True):
        l = sorted(d, key=lambda x: x["fgap_median"])

        best = l[0]["fgap"]

        # _, pval = scipy.stats.levene(*[e["fgap"] for e in l])

        t = PrettyTable()
        t.title = scope["problem"]
        t.field_names = ['Algorithm', 'fgap_median', 'fgap_mean', 'fgap_std', 'shapiro', 'levene', 't-test', 'wilcoxon']

        for i, e in enumerate(l):
            f = e["fgap"]

            _, pval = scipy.stats.shapiro(f)
            shapiro = "*" if pval >= 0.01 else ""

            _, pval = scipy.stats.levene(best, f)
            levene = "* (%.3f)" % pval if pval >= 0.05 else ""

            _, pval = scipy.stats.ttest_ind(f, best, alternative="greater")
            ttest = "* (%.3f)" % pval if pval >= 0.05 else ""

            if len(best) == len(f):
                _, pval = scipy.stats.wilcoxon(f, best, zero_method="zsplit", alternative="greater")
                wilcoxon = "* (%.3f)" % pval if pval >= 0.05 else ""
            else:
                wilcoxon = "x"

            t.add_row(
                [e["algorithm"], "%.10f" % e["fgap_median"], "%.10f" % e["fgap_mean"], "%.10f" % e["fgap_std"], shapiro,
                 levene, ttest, wilcoxon])

        print(t)
        print()

    conv = ConvergenceAnalyzer().run(results, group_by=["problem", "algorithm"])

    attrs = {"conv_gap": "np.median", "n_evals": "first"}
    median = GroupBy(attrs).run(conv, group_by=["problem", "algorithm"])

    plot = False

    if plot:

        for scope, d in filter_by(median, ["problem"], return_group=True):

            plt.figure()
            plt.title(scope["problem"])
            plt.yscale("log")

            for entry in d:
                plt.plot(entry["n_evals"], entry["conv_gap_mean"], label=entry["algorithm"])

            plt.legend()
            plt.show()

    ranks = {}

    for scope, d in filter_by(mean, ["problem"], return_group=True):

        A = [e["algorithm"] for e in d]
        perf = np.array([e["fgap_median"] for e in d])
        rank = np.argsort(np.argsort(perf))

        for a, r in zip(A, rank):

            if a not in ranks:
                ranks[a] = []

            ranks[a].append(r + 1)

    cnt = np.array([len(v) for k, v in ranks.items()])
    assert cnt.min() == cnt.max(), "All algorithms must have been run on all problems!"
    k = cnt.min()

    R = np.column_stack([v for k, v in ranks.items()])
    R_total = R.mean(axis=0)

    t = PrettyTable(['problem'] + list(ranks.keys()))

    p = list(benchmark.problems.keys())

    for i, row in enumerate(R):
        t.add_row([p[i]] + list(row))

    t.add_row(["Total"] + [np.round(e, 3) for e in R_total])

    print(t)



from pymoo.algorithms.moo.nsga2 import RankAndCrowdingSurvival
from pymoo.algorithms.soo.nonconvex.de import DE
from pymoo.core.population import Population
from pymoo.docs import parse_doc_string
from pymoo.operators.mutation.pm import PM
from pymoo.util.display import MultiObjectiveDisplay
from pymoo.util.dominator import get_relation
from pymoo.util.termination.default import MultiObjectiveDefaultTermination


class GDE3(DE):

    def __init__(self,
                 CR=0.3,
                 F=None,
                 variant="DE/rand/1/bin",
                 mutation=PM(eta=15),
                 **kwargs):

        super().__init__(CR=CR, F=F, variant=variant, mutation=mutation, display=MultiObjectiveDisplay(), **kwargs)
        self.default_termination = MultiObjectiveDefaultTermination()

    def _initialize_advance(self, infills=None, **kwargs):
        pass

    def _advance(self, infills=None, **kwargs):
        assert infills is not None, "This algorithms uses the AskAndTell interface thus 'infills' must to be provided."

        # the indices for the creating of offsprings considered in the last generation
        I = infills.get("index")

        # the individuals that are considered for the survival later and final survive
        survivors = []

        # now for each of the infill solutions
        for k, i in enumerate(I):

            # get the offspring an the parent it is coming from
            off, parent = infills[k], self.pop[i]

            # check whether the new solution dominates the parent or not
            rel = get_relation(parent, off)

            # if indifferent we add both
            if rel == 0:
                survivors.extend([parent, off])

            # if offspring dominates parent
            elif rel == -1:
                survivors.append(off)

            # if parent dominates offspring
            else:
                survivors.append(parent)

        # create the population
        survivors = Population.create(*survivors)

        # perform a survival to reduce to pop size
        self.pop[I] = RankAndCrowdingSurvival().do(self.problem, survivors, n_survive=len(I))


parse_doc_string(GDE3.__init__)



from pymoo.core.problem import calc_constr
from pymoo.problems.meta import MetaProblem
from pymoo.util.misc import from_dict


class ConstraintViolationAsObjective(MetaProblem):

    def __init__(self, problem, eps=1e-6):
        super().__init__(problem)
        self.n_obj = 1
        self.n_constr = 0
        self.eps = eps

    def do(self, x, out, *args, **kwargs):
        super().do(x, out, *args, **kwargs)

        F, G = from_dict(out, "F", "G")

        assert G is not None, "To converge a function's constraint to objective it needs G to be set!"

        out["__F__"] = out["F"]
        out["__G__"] = out["G"]

        out["F"] = calc_constr(G, eps=self.eps, beta=1.0)
        del out["G"]



import numpy as np

from pymoo.algorithms.base.local import LocalSearch
from pymoo.docs import parse_doc_string
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.core.replacement import is_better
from pymoo.operators.repair.to_bound import set_to_bounds_if_outside_by_problem
from pymoo.util.display import SingleObjectiveDisplay
from pymoo.util.optimum import filter_optimum


# =========================================================================================================
# Implementation
# =========================================================================================================


class PatternSearchDisplay(SingleObjectiveDisplay):

    def __init__(self, **kwargs):
        super().__init__(favg=False, **kwargs)

    def _do(self, problem, evaluator, algorithm):
        super()._do(problem, evaluator, algorithm)
        self.output.append("rho", algorithm._rho)


class PatternSearch(LocalSearch):
    def __init__(self,
                 init_delta=0.25,
                 rho=0.5,
                 step_size=1.0,
                 display=PatternSearchDisplay(),
                 **kwargs):
        """
        An implementation of well-known Hooke and Jeeves Pattern Search.

        Parameters
        ----------

        x0 : numpy.array
            The initial value where the local search should be initiated. If not provided `n_sample_points` are created
            created using latin hypercube sampling and the best solution found is set to `x0`.

        n_sample_points : int
            Number of sample points to be used to determine the initial search point. (Only used of `x0` is not provided)

        delta : float
            The `delta` values which is used for the exploration move. If lower and upper bounds are provided the
            value is in relation to the overall search space. For instance, a value of 0.25 means that initially the
            pattern is created in 25% distance of the initial search point.

        rho : float
            If the move was unsuccessful then the `delta` value is reduced by multiplying it with the value provided.
            For instance, `explr_rho` implies that with a value of `delta/2` is continued.

        step_size : float
            After the exploration move the new center is determined by following a promising direction.
            This value defines how large to step on this direction will be.

        """

        super().__init__(display=display, **kwargs)
        self.rho = rho
        self.step_size = step_size
        self.init_delta = init_delta

        self._rho = rho
        self._delta = None
        self._center = None
        self._current = None
        self._trial = None
        self._direction = None
        self._sign = None

    def _setup(self, problem, **kwargs):

        if problem.has_bounds():
            xl, xu = problem.bounds()
            self._delta = self.init_delta * (xu - xl)
        else:
            self._delta = np.abs(self.x0) / 2.0
            self._delta[self._delta <= 1.0] = 1.0

    def _initialize_advance(self, infills=None, **kwargs):
        super()._initialize_advance(infills, **kwargs)
        self._center, self._explr = self.x0, None
        self._sign = np.ones(self.problem.n_var)

    def _local_advance(self, **kwargs):

        # this happens only in the first iteration
        if self._explr is None:
            self._explr = self._exploration_move(self._center)

        else:
            # whether the last iteration has resulted in a new optimum
            has_improved = is_better(self._explr, self._center, eps=0.0)

            # that means that the exploration did not found any new point and was thus unsuccessful
            if not has_improved:

                # keep track of the rho values in the normalized space
                self._rho = self._rho * self.rho

                # explore around the current center to try finding a suitable direction
                self._explr = self._exploration_move(self._center)

            # if we have found a direction in the last iteration to be worth following
            else:

                # get the direction which was successful in the last move
                self._direction = (self._explr.X - self._center.X)

                # declare the exploration point the new center
                self._center = self._explr

                # use the pattern move to get a new trial vector along that given direction
                self._trial = self._pattern_move(self._center, self._direction)

                # get the delta sign adjusted for the exploration
                self._sign = calc_sign(self._direction)

                # explore around the current center to try finding a suitable direction
                self._explr = self._exploration_move(self._trial)

        self.pop = Population.create(self._center, self._explr)

    def _exploration_move(self, center):

        # randomly iterate over all variables
        for k in range(self.problem.n_var):

            # the value to be tried first is given by the amount times the sign
            _delta = self._sign[k] * self._rho * self._delta

            # make a step of delta on the k-th variable
            _explr = step(self.problem, center.X, _delta, k)
            self.evaluator.eval(self.problem, _explr, algorithm=self)

            if is_better(_explr, center, eps=0.0):
                center = _explr

            # if not successful try the other direction
            else:

                # now try the negative value of delta and see if we can improve
                _explr = step(self.problem, center.X, -1 * _delta, k)
                self.evaluator.eval(self.problem, _explr, algorithm=self)

                if is_better(_explr, center, eps=0.0):
                    center = _explr

        return center

    def _pattern_move(self, _current, _direction):

        # calculate the new X and repair out of bounds if necessary
        X = _current.X + self.step_size * _direction
        set_to_bounds_if_outside_by_problem(self.problem, X)

        # create the new center individual and evaluate it
        trial = Individual(X=X)
        self.evaluator.eval(self.problem, trial, algorithm=self)

        return trial

    def _set_optimum(self):
        pop = self.pop if self.opt is None else Population.merge(self.opt, self.pop)
        self.opt = filter_optimum(pop, least_infeasible=True)


def calc_sign(direction):
    sign = np.sign(direction)
    sign[sign == 0] = -1
    return sign


def step(problem, x, delta, k):
    # copy and add delta to the new point
    X = np.copy(x)

    # if the problem has bounds normalize the delta
    # if problem.has_bounds():
    #     xl, xu = problem.bounds()
    #     delta *= (xu[k] - xl[k])

    # now add to the current solution
    X[k] = X[k] + delta[k]

    # repair if out of bounds if necessary
    X = set_to_bounds_if_outside_by_problem(problem, X)

    return Individual(X=X)


parse_doc_string(PatternSearch.__init__)



import numpy as np


def crossover_mask(X, M):
    # convert input to output by flatting along the first axis
    _X = np.copy(X)
    _X[0][M] = X[1][M]
    _X[1][M] = X[0][M]
    return _X


def row_at_least_once_true(M):
    _, d = M.shape
    for k in np.where(~np.any(M, axis=1))[0]:
        M[k, np.random.randint(d)] = True
    return M



import numpy as np

from pymoo.util.misc import list_of_dicts_unique, list_of_dicts_filter


def filter_by(data, filters, return_group=False):
    yield from filter_by_rec(data, filters, {}, return_scope=return_group)


def filter_by_rec(data, filters, scope, return_scope):
    if len(filters) == 0:
        if return_scope:
            yield scope, data
        else:
            yield data
    else:
        entry = filters[0]

        if isinstance(entry, tuple):
            k, vals = entry
        else:
            k = entry
            vals = list_of_dicts_unique(data, k)

        for v in vals:
            _data = list_of_dicts_filter(data, (k, v))
            _filters = filters[1:]
            _scope = {**dict(scope), **{k: v}}
            yield from filter_by_rec(_data, _filters, _scope, return_scope)


def fill_forward_if_nan(vals):
    current = np.nan
    for k in range(len(vals)):

        if not np.isnan(vals[k]):
            current = vals[k]

        if np.isnan(vals[k]):
            vals[k] = current
    return vals


def at_least2d(x, expand="c"):
    if x.ndim == 1:
        if expand == "c":
            return x[:, None]
        elif expand == "r":
            return x[None, :]
    else:
        return x



from abc import abstractmethod

import numpy as np

from pymoo.core.population import Population


# ---------------------------------------------------------------------------------------------------------
# Survival
# ---------------------------------------------------------------------------------------------------------


class Survival:

    def __init__(self, filter_infeasible=True):
        super().__init__()
        self.filter_infeasible = filter_infeasible

    def do(self,
           problem,
           pop,
           *args,
           n_survive=None,
           return_indices=False,
           **kwargs):

        # make sure the population has at least one individual
        if len(pop) == 0:
            return pop

        if n_survive is None:
            n_survive = len(pop)

        n_survive = min(n_survive, len(pop))

        # if the split should be done beforehand
        if self.filter_infeasible and problem.n_constr > 0:

            # split feasible and infeasible solutions
            feas, infeas = split_by_feasibility(pop, eps=0.0, sort_infeasbible_by_cv=True)

            if len(feas) == 0:
                survivors = Population()
            else:
                survivors = self._do(problem, pop[feas], *args, n_survive=min(len(feas), n_survive), **kwargs)

            # calculate how many individuals are still remaining to be filled up with infeasible ones
            n_remaining = n_survive - len(survivors)

            # if infeasible solutions needs to be added
            if n_remaining > 0:
                survivors = Population.merge(survivors, pop[infeas[:n_remaining]])

        else:
            survivors = self._do(problem, pop, *args, n_survive=n_survive, **kwargs)

        if return_indices:
            H = {}
            for k, ind in enumerate(pop):
                H[ind] = k
            return [H[survivor] for survivor in survivors]
        else:
            return survivors

    @abstractmethod
    def _do(self, problem, pop, *args, n_survive=None, **kwargs):
        pass


def split_by_feasibility(pop, eps=0.0, sort_infeasbible_by_cv=True):
    CV = pop.get("CV")

    b = (CV <= eps)

    feasible = np.where(b)[0]
    infeasible = np.where(~b)[0]

    if sort_infeasbible_by_cv:
        infeasible = infeasible[np.argsort(CV[infeasible, 0])]

    return feasible, infeasible


def calc_adapt_eps(pop):
    cv = pop.get("CV")[:, 0]
    cv_mean = np.median(cv)
    fr = (cv <= 0).sum() / len(cv)
    return cv_mean * fr



from pymoo.problems.meta import MetaProblem


class MultiToSingleObjective(MetaProblem):

    def __init__(self, problem, decomposition, kwargs=None):
        super().__init__(problem)
        self.decomposition = decomposition
        self.kwargs = kwargs if not None else dict()

        self.n_obj = 1

    def do(self, x, out, *args, **kwargs):
        super().do(x, out, *args, **kwargs)

        F = out["F"]
        out["__F__"] = F
        out["F"] = self.decomposition.do(F, **self.kwargs)[:, None]







from pymoo.algorithms.soo.convex.deriv.newton import NewtonMethod
from pymoo.factory import Himmelblau
from pymoo.optimize import minimize
from pymoo.problems.autodiff import AutomaticDifferentiation

problem = AutomaticDifferentiation(Himmelblau())

algorithm = NewtonMethod()

res = minimize(problem,
               algorithm,
               seed=2,
               verbose=True)

print(res.X)
print(res.F)



import numpy as np

from pymoo.experimental.benchmarking.analyzer.analyzer import Analyzer
from pymoo.util.misc import from_dict


class SingleObjectiveAnalyzer(Analyzer):

    def do(self, data, benchmark=None, inplace=False, **kwargs):
        assert benchmark is not None, "The benchmark is necessary to retrieve the known optimum of a funtion"

        problem = benchmark.problems[data["problem"]]["obj"]
        CV, F = from_dict(data, "CV", "F")

        f = F[0, 0]
        cv = CV[0, 0]

        fopt = problem.pareto_front()
        if fopt is not None:
            fopt = fopt[0, 0].astype(np.float)

        fgap = f - fopt

        ret = {
            "f": f,
            "cv": cv,
            "fopt": fopt,
            "fgap": fgap
        }

        if inplace:
            for k, v in ret.items():
                data[k] = v

        return ret






import numpy as np

from pymoo.core.crossover import Crossover
from pymoo.core.population import Population
from pymoo.operators.crossover.util import row_at_least_once_true


def mut_binomial(n, m, prob, at_least_once=True):
    M = np.random.random((n, m)) < prob

    if at_least_once:
        M = row_at_least_once_true(M)

    return M


class BinomialCrossover(Crossover):

    def __init__(self, bias, **kwargs):
        super().__init__(2, 1, **kwargs)
        self.bias = bias

    def do(self, problem, pop, parents, **kwargs):
        X = pop.get("X")[parents.T].copy()

        _, n_matings, n_var = X.shape
        M = mut_binomial(n_matings, n_var, self.bias, at_least_once=True)

        Xp = X[0]
        Xp[~M] = X[1][~M]

        return Population.new(X=Xp)



import autograd.numpy as anp

from pymoo.problems.meta import MetaProblem


class BoundariesAsConstraints(MetaProblem):

    def __init__(self, problem):
        super().__init__(problem)
        self.n_constr = self.n_constr + 2 * self.n_var

    def do(self, X, out, *args, **kwargs):
        self.problem.do(X, out, *args, **kwargs)

        # get the boundaries for normalization
        xl, xu = self.bounds()

        # add the boundary constraint if enabled
        _G = anp.column_stack([xl-X, X-xu])

        out["G"] = anp.column_stack([out["G"], _G]) if out.get("G") is not None else _G

        if "dG" in out:
            _dG = anp.zeros((len(X), 2 * self.n_var, self.n_var))
            _dG[:, :self.n_var, :] = - anp.eye(self.n_var)
            _dG[:, self.n_var:, :] = anp.eye(self.n_var)
            out["dG"] = anp.column_stack([out["dG"], _dG]) if out.get("dG") is not None else _dG






from abc import abstractmethod
from copy import deepcopy

from pymoo.algorithms.soo.nonconvex.ga import FitnessSurvival, GA
from pymoo.docs import parse_doc_string
from pymoo.core.algorithm import Algorithm
from pymoo.core.evaluator import set_cv, set_feasibility, Evaluator
from pymoo.core.individual import Individual
from pymoo.core.population import Population
from pymoo.problems.single.traveling_salesman import create_random_tsp_problem
from pymoo.util.display import SingleObjectiveDisplay
from pymoo.util.roulette import RouletteWheelSelection


# =========================================================================================================
# Display
# =========================================================================================================


class ACODisplay(SingleObjectiveDisplay):

    def _do(self, problem, evaluator, algorithm):
        super()._do(problem, evaluator, algorithm)

        F = algorithm.off.get("F")
        self.output.append("it-best", F.min(), width=12)
        self.output.append("it-avg", F.mean(), width=12)


# =========================================================================================================
# Entry
# =========================================================================================================


class Entry:

    def __init__(self, key=None, heuristic=None, pheromone=None) -> None:
        super().__init__()
        self.key = key
        self.heuristic = heuristic
        self.pheromone = pheromone


# =========================================================================================================
# Ant
# =========================================================================================================

class Ant(Individual):

    def __init__(self,
                 alpha=0.1,
                 beta=2.0,
                 q0=None,
                 **kwargs) -> None:
        """


        Parameters
        ----------

        alpha : float
            Relative importance of pheromones

        beta : float
            Relative importance of heuristic information

        q0 : float
            Probability of choosing the maximum values according to selection metric and not doing
            the roulette wheel selection. For Ant Colony System 0.9 was proposed.

        """
        super().__init__(**kwargs)

        self.alpha = alpha
        self.beta = beta
        self.q0 = q0

        self.path = []
        self.data = {}

        self.problem = None
        self.pheromones = None

    def initialize(self, problem, pheromones):
        self.problem = problem
        self.pheromones = pheromones

    def next(self):
        candidates = self._next()

        # information coming from the pheromones
        tau = []
        for entry in candidates:
            key = entry.key

            # the first time this key is accessed it is initialized before - if sparse implementation is used
            if not self.pheromones.has(key):
                self.pheromones._initialize(key)

            # get the pheromones on the field and add it to the list
            _tau = self.pheromones.get(key)
            tau.append(_tau)

        tau = np.array(tau)

        # information coming from the heuristics
        eta = np.array([e.heuristic for e in candidates])

        # p = tau ** self.alpha + eta ** self.beta
        p = tau + eta ** self.beta
        p = p / p.sum()

        if self.q0 is not None and np.random.random() <= self.q0:
            k = p.argmax()
        else:
            k = RouletteWheelSelection(p).next()

        entry = candidates[k]
        self.notify(entry)

    def has_next(self):
        return self._has_next()

    def last(self):
        if len(self.path) > 0:
            return self.path[-1]
        else:
            return None

    def notify(self, entry):
        self.path.append(entry)
        self._notify(entry)

    def get_values(self, key="key", as_numpy_array=True):
        ret = [e.__dict__[key] for e in self.path]
        if as_numpy_array:
            ret = np.array(ret)
        return ret

    def finalize(self):
        self._finalize()

    @abstractmethod
    def _next(self):
        raise Exception("To be implemented")

    @abstractmethod
    def _has_next(self):
        raise Exception("To be implemented")

    def _finalize(self):
        pass

    def _notify(self, entry):
        pass


# =========================================================================================================
# Pheromones
# =========================================================================================================


def get_symmetric(key):
    if isinstance(key, tuple) or isinstance(key, tuple):
        if len(key) == 2:
            return min(key), max(key)

    raise Exception("Symmetric pheromones can only be used for a two-dimensional matrix.")


class Pheromones:

    def __init__(self,
                 tau_init=0.1,
                 rho=0.1,
                 symmetric=True
                 ) -> None:
        """

        Parameters
        ----------
        rho : float
            The evaporation parameter.

        symmetric : bool
            Whether the pheromones should be set symmetrically

        """
        super().__init__()
        self.tau_init = tau_init
        self.tau_init_evaporated = tau_init
        self.rho = rho
        self.symmetric = symmetric

        self.data = {}

    def initialize(self, key):
        self.set(key, self.tau_init_evaporated)

    def _get(self, key):
        key = get_symmetric(key) if self.symmetric else key
        return key

    def _set(self, key, value):
        key = get_symmetric(key) if self.symmetric else key
        self.data[key] = value

    def get(self, key):
        return self.data.get(self._get(key))

    def has(self, key):
        return self._get(key) in self.data

    def set(self, key, value):
        self._set(key, value)

    def evaporate(self):
        keys = list(self.data.keys())

        for key in keys:
            value = self.data[self._get(key)]
            self._set(key, value * (1 - self.rho))

        self.tau_init_evaporated = self.tau_init_evaporated * (1 - self.rho)

    def update(self, key, value):
        self.data[self._get(key)] += value


# =========================================================================================================
# Mock Evaluator
# =========================================================================================================


class MockEvaluator(Evaluator):

    def __init__(self, **kwargs):
        super().__init__(skip_already_evaluated=False, **kwargs)

    def _eval(self, problem, pop, **kwargs):
        pass


# =========================================================================================================
# Algorithm
# =========================================================================================================


class ACO(Algorithm):

    def __init__(self,
                 ant,
                 pheromones,
                 n_ants=10,
                 global_update='best',
                 local_update=True,
                 evaluate_each_ant=True,
                 display=ACODisplay(),
                 **kwargs):
        """

        Parameters
        ----------

        ant : class
            An objective defining the behavior of each ant

        pheromones : class
            The pheromone implementation storing the amount of pheromones on each edge,
            the evaporation and update.

        n_ants : int
            Number of ants to be used each iteration

        global_update : {'all', 'it-best', 'best'}

        """

        super().__init__(display=display, **kwargs)

        # make the ant always to be a function being able to call
        if not callable(ant):
            proto = deepcopy(ant)

            def new_ant():
                return deepcopy(proto)

            ant = new_ant

        self.ant = ant
        self.pheromones = pheromones
        self.n_ants = n_ants
        self.global_update = global_update
        self.local_update = local_update
        self.evaluate_each_ant = evaluate_each_ant

    def _initialize_infill(self):
        self.evaluator = Evaluator(skip_already_evaluated=False) if self.evaluate_each_ant else MockEvaluator()
        self.opt = Population()
        self.step()

    def step(self):

        # initialize all ants to be used in this iteration
        ants = []
        for k in range(self.n_ants):
            ant = self.ant()
            ant._initialize(self.problem, self.pheromones)
            ants.append(ant)

        active = list(range(self.n_ants))

        while len(active) > 0:

            for k in active:
                ant = ants[k]

                if ant.has_next():
                    ant.next()

                    if self.local_update:
                        e = ant.last()
                        if e is None or e.pheromone is None:
                            raise Exception("For a local update the ant has to set the pheromones when notified.")
                        else:
                            self.pheromones.set(e.key, self.pheromones.get(e.key) * ant.alpha + e.pheromone * ant.alpha)
                            # self.pheromones.update(e.key, e.pheromone * ant.alpha)

                else:
                    ant.finalize()
                    active = [i for i in active if i != k]

        colony = Population.create(*ants)

        # this evaluation can be disabled or faked if evaluate_each_ant is false - then the finalize method of the
        # ant has to set the objective and/or constraint values accordingly
        self.evaluator.eval(self.problem, colony)
        set_cv(colony)
        set_feasibility(colony)

        # set the current best including the new colony
        opt = FitnessSurvival().do(problem, Population.merge(colony, self.opt), 1)

        # do the evaporation after this iteration
        self.pheromones.evaporate()

        # select the ants to be used for the global pheromone update
        if self.global_update == "all":
            ants_to_update = colony
        elif self.global_update == "it-best":
            ants_to_update = FitnessSurvival().do(problem, colony, 1)
        elif self.global_update == "best":
            ants_to_update = self.opt
        else:
            raise Exception("Unknown value for global updating the pheromones!")

        # now spread the pheromones for each ant depending on performance
        for ant in ants_to_update:
            for e in ant.path:
                if e.pheromone is None:
                    raise Exception("The ant has to set the pheromone of each entry in the path.")
                else:
                    self.pheromones.update(e.key, e.pheromone * pheromones.rho)

        self.pop, self.off = colony, colony
        self.opt = opt

    def _set_optimum(self, **kwargs):
        pass


parse_doc_string(ACO.__init__)


# =========================================================================================================
# TSP Example
# =========================================================================================================


class GraphAnt(Ant):

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.not_visited = None
        self.costs = 0.0

    def initialize(self, *args, **kwargs):
        super().initialize(*args, **kwargs)
        self.not_visited = np.full(self.problem.n_var, True)

    def _notify(self, entry):
        self.not_visited[list(entry.key)] = False

        a, b = entry.key
        self.costs += self.problem.D[a, b]

    def heuristic(self, key):
        _current, _next = key
        _heur = 1 / self.problem.D[_current, _next]
        return _heur

    def _next(self):
        ret = []

        if len(self.path) == 0:
            _current = np.random.randint(self.problem.n_var)
            self.not_visited[_current] = False
        else:
            _, _current = self.path[-1].key

        remaining = np.where(self.not_visited)[0]

        if len(remaining) == 0:
            return_to_first, _ = self.path[0].key
            remaining = [return_to_first]

        for e in remaining:
            _next = int(e)
            _key = (_current, _next)

            # the heuristic value for this edge - default 1 / cost(i,j)
            _heur = self.heuristic(_key)

            # if local pheromone updates are enabled this needs to be set here
            _pheromone = self.pheromones.tau_init

            entry = Entry(key=_key, heuristic=_heur, pheromone=_pheromone)
            ret.append(entry)

        return ret

    def _finalize(self):

        self.X = np.array([e.key[0] for e in self.path])
        assert (len(self.X) == self.problem.n_var)

        self.F = np.array([self.costs])

        for entry in self.path:
            entry.pheromone = 1 / self.costs

    def _has_next(self):
        return len(self.path) < self.problem.n_var


def solve_nearest_neighbor(tsp, return_time=False):
    n_cities = tsp.n_var
    time = 0.0

    # initialize the array to store what cities have been visited
    visited = np.full(n_cities, False)

    # randomly choose the first city to start
    _start = np.random.randint(n_cities)
    visited[_start] = True

    # create the initial tour
    tour = [_start]

    # until we have not visited all cities
    while len(tour) < n_cities:
        _current = tour[-1]

        remaining = np.where(~visited)[0]
        d = tsp.D[_current][remaining]

        k = d.argmin()
        closest = remaining[k]
        visited[closest] = True
        time += d[k]

        tour.append(closest)

    first, last = tour[0], tour[-1]
    time += tsp.D[last][first]

    tour = np.array(tour)

    if not return_time:
        return tour
    else:
        return tour, time


if __name__ == "__main__":
    problem = create_random_tsp_problem(50, 100, seed=1)

    import numpy as np
    from pymoo.core.repair import Repair


    class StartFromZeroRepair(Repair):

        def _do(self, problem, pop, **kwargs):
            X = pop.get("X")
            I = np.where(X == 0)[1]

            for k in range(len(X)):
                i = I[k]
                x = X[k]
                _x = np.concatenate([x[i:], x[:i]])
                pop[k].set("X", _x)

            return pop



    from pymoo.optimize import minimize
    from pymoo.factory import get_crossover, get_mutation, get_sampling
    from pymoo.problems.single.traveling_salesman import create_random_tsp_problem
    from pymoo.util.termination.default import SingleObjectiveDefaultTermination

    algorithm = GA(
        pop_size=20,
        sampling=get_sampling("perm_random"),
        crossover=get_crossover("perm_erx"),
        mutation=get_mutation("perm_inv"),
        repair=StartFromZeroRepair(),
        eliminate_duplicates=True
    )

    # if the algorithm did not improve the last 200 generations then it will terminate (and disable the max generations)
    termination = SingleObjectiveDefaultTermination(n_last=200, n_max_gen=np.inf)

    # res = minimize(
    #     problem,
    #     algorithm,
    #     termination,
    #     seed=1,
    # )
    #
    # print("Traveling Time:", np.round(res.F[0], 3))
    # print("Function Evaluations:", res.algorithm.evaluator.n_eval)

    ant = lambda: GraphAnt(alpha=0.1, beta=2.0, q0=0.9)

    nn_times = []
    for k in range(10):
        _, time = solve_nearest_neighbor(problem, return_time=True)
        nn_times.append(time)
    time = min(nn_times)

    tau_init = (problem.n_var * time) ** -1

    pheromones = Pheromones(tau_init=tau_init, rho=0.1)

    for i in range(problem.n_var):
        for j in range(i + 1, problem.n_var):
            pheromones.initialize((i, j))

    algorithm = ACO(ant, pheromones, n_ants=10, global_update="best", local_update=True)

    minimize(problem,
             algorithm,
             ("n_gen", 300),
             seed=1,
             verbose=True)
